import six
from mdt.components_loader import ComponentConfig, ComponentBuilder, method_binding_meta
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.parameters import StaticMapParameter, ProtocolParameter, ModelDataParameter, \
    FreeParameter
from mot.model_building.parameter_functions.priors import UniformWithinBoundsPrior
from mot.model_building.parameter_functions.proposals import GaussianProposal
from mot.model_building.parameter_functions.sample_statistics import GaussianPSS
from mot.model_building.parameter_functions.transformations import IdentityTransform

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterConfig(ComponentConfig):
    """The cascade config to inherit from.

    These configs are loaded on the fly by the ParametersBuilder

    Config options:
        name (str): the name of the parameter, defaults to the class name
        description (str): the description of this parameter
        data_type (str or DataType): either a string we use as datatype or the actual datatype itself
        type (str): the type of parameter (free, protocol or model_data)
    """
    name = ''
    description = ''
    data_type = 'mot_float_type'
    type = None


class ProtocolParameterConfig(ParameterConfig):
    """The default config options for protocol parameters.

    This sets the attribute type to protocol.
    """
    type = 'protocol'
    data_type = 'mot_float_type'


class FreeParameterConfig(ParameterConfig):
    """The default config options for free parameters.

    This sets the attribute type to free.

    Attributes:
        init_value (float): the initial value
        fixed (boolean or ndarray of float): if this parameter is fixed or not. If not fixed this should
            hold a reference to a value or a matrix
        lower_bound (float): the lower bounds
        upper_bound (float): the upper bounds
        parameter_transform: the parameter transformation
        sampling_proposal: the proposal function
        sampling_prior: the prior function
        sampling_statistics: the sampling statistic, used after the sampling
    """
    type = 'free'
    data_type = 'mot_float_type'
    fixed = False
    init_value = 0.03
    lower_bound = 0.0
    upper_bound = 4.0
    parameter_transform = IdentityTransform()
    sampling_proposal = GaussianProposal(1.0)
    sampling_prior = UniformWithinBoundsPrior()
    sampling_statistics = GaussianPSS()


class ModelDataParameterConfig(ParameterConfig):
    """The default config options for model data parameters.

    This sets the attribute type to model_data.
    """
    type = 'model_data'
    value = None


class StaticMapParameterConfig(ParameterConfig):
    """The default config options for static data parameters.

    This sets the attribute type to static_map.
    """
    type = 'static_map'
    value = None


class ParameterBuilder(ComponentBuilder):

    def create_class(self, template):
        """Creates classes with as base class DMRICompositeModel

        Args:
            template (ParameterConfig): the configuration for the parameter.
        """
        data_type = template.data_type
        if isinstance(data_type, six.string_types):
            data_type = CLDataType.from_string(data_type)

        if template.type.lower() == 'protocol':
            class AutoProtocolParameter(method_binding_meta(template, ProtocolParameter)):
                def __init__(self):
                    super(AutoProtocolParameter, self).__init__(data_type, template.name)
            return AutoProtocolParameter

        elif template.type.lower() == 'free':
            class AutoFreeParameter(method_binding_meta(template, FreeParameter)):
                def __init__(self):
                    super(AutoFreeParameter, self).__init__(
                        data_type,
                        template.name,
                        template.fixed,
                        template.init_value,
                        template.lower_bound,
                        template.upper_bound,
                        parameter_transform=template.parameter_transform,
                        sampling_proposal=template.sampling_proposal,
                        sampling_prior=template.sampling_prior,
                        sampling_statistics=template.sampling_statistics
                    )
            return AutoFreeParameter

        elif template.type.lower() == 'model_data':
            class AutoModelDataParameter(method_binding_meta(template, ModelDataParameter)):
                def __init__(self):
                    super(AutoModelDataParameter, self).__init__(data_type, template.name, template.value)
            return AutoModelDataParameter

        elif template.type.lower() == 'static_map':
            class AutoStaticMapParameter(method_binding_meta(template, StaticMapParameter)):
                def __init__(self):
                    super(AutoStaticMapParameter, self).__init__(data_type, template.name, template.value)
            return AutoStaticMapParameter



import six
from mdt.components_loader import CompartmentModelsLoader
from .CompositeModelExpression import CompositeModelExpressionSemantics, CompositeModelExpressionParser


class Semantics(CompositeModelExpressionSemantics):

    def __init__(self):
        super(Semantics, self).__init__()
        self._compartments_loader = CompartmentModelsLoader()

    def expr(self, ast):
        if not isinstance(ast, list):
            return ast
        if isinstance(ast, list):
            return ast[0], ast[2], ast[1]
        return ast

    def term(self, ast):
        if not isinstance(ast, list):
            return ast
        if isinstance(ast, list):
            return ast[0], ast[2], ast[1]
        return ast

    def factor(self, ast):
        if isinstance(ast, list):
            return ast[1]
        return ast

    def model(self, ast):
        if isinstance(ast, six.string_types):
            return self._compartments_loader.load(ast)
        else:
            return self._compartments_loader.load(ast[0], ast[2])


def parse(model_expression):
    """Parse the given model expression into a suitable model tree.

    Args:
        model_expression (str): the model expression string. Example:

        .. code-block:: none

            S0 * ( (Weight(Wball) * Ball) +
                   (Weight(Wstick) * Stick ) )

        If the model name is followed by parenthesis the string in parenthesis will represent the model's nickname.

    Returns:
        :class:`list`: the compartment model tree for use in composite models.
    """
    parser = CompositeModelExpressionParser(parseinfo=False)
    return parser.parse(model_expression, rule_name='result', semantics=Semantics())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'runtime_settings_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_RuntimeSettingsDialog(object):
    def setupUi(self, RuntimeSettingsDialog):
        RuntimeSettingsDialog.setObjectName("RuntimeSettingsDialog")
        RuntimeSettingsDialog.resize(844, 243)
        self.verticalLayout = QtWidgets.QVBoxLayout(RuntimeSettingsDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(RuntimeSettingsDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(RuntimeSettingsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(RuntimeSettingsDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.cldevicesSelection = QtWidgets.QListWidget(RuntimeSettingsDialog)
        self.cldevicesSelection.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.cldevicesSelection.setObjectName("cldevicesSelection")
        self.gridLayout.addWidget(self.cldevicesSelection, 0, 1, 1, 1)
        self.label_10 = QtWidgets.QLabel(RuntimeSettingsDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 0, 2, 1, 1)
        self.label = QtWidgets.QLabel(RuntimeSettingsDialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(RuntimeSettingsDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(RuntimeSettingsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(RuntimeSettingsDialog)
        self.buttonBox.accepted.connect(RuntimeSettingsDialog.accept)
        self.buttonBox.rejected.connect(RuntimeSettingsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(RuntimeSettingsDialog)

    def retranslateUi(self, RuntimeSettingsDialog):
        _translate = QtCore.QCoreApplication.translate
        RuntimeSettingsDialog.setWindowTitle(_translate("RuntimeSettingsDialog", "Runtime settings"))
        self.label_3.setText(_translate("RuntimeSettingsDialog", "Runtime settings"))
        self.label_4.setText(_translate("RuntimeSettingsDialog", "Runtime settings for all compute operations."))
        self.label_10.setText(_translate("RuntimeSettingsDialog", "(Select the devices you would like to use)"))
        self.label.setText(_translate("RuntimeSettingsDialog", "OpenCL devices:"))




import glob
import os
import mdt
from mdt.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'HCP MGH',
             'description': 'The profile for the MGH data from the Human Connectome project',
             'directory_layout':
'''
This assumes that you downloaded and extracted the MGH data in one folder and that you now have one folder per subject.

You can provide the noise standard deviation to use using a noise_std file containing a single float.

Example directory layout:
    /mgh_*/diff/preproc/mri/diff_preproc.nii(.gz)
    /mgh_*/diff/preproc/bvals.txt
    /mgh_*/diff/preproc/bvecs_fsl_moco_norm.txt

Optional items (these will take precedence if present):
    /mgh_*/diff/preproc/diff_preproc.bval
    /mgh_*/diff/preproc/diff_preproc.bvec
    /mgh_*/diff/preproc/diff_preproc.prtcl
    /mgh_*/diff/preproc/diff_preproc_mask.nii(.gz)
    /mgh_*/diff/preproc/mri/diff_preproc_mask.nii(.gz)
    /mgh_*/diff/preproc/noise_std.{txt,nii,nii.gz}
'''}


class HCP_MGH(SimpleBatchProfile):

    def __init__(self):
        super(HCP_MGH, self).__init__()
        self._output_base_dir = 'diff/preproc/output'

    def _get_subjects(self):
        dirs = sorted([os.path.basename(f) for f in glob.glob(os.path.join(self._root_dir, '*'))])
        subjects = []
        for subject_id in dirs:
            pjoin = mdt.make_path_joiner(self._root_dir, subject_id, 'diff', 'preproc')
            if os.path.isdir(pjoin()):
                dwi_fname = list(glob.glob(pjoin('mri', 'diff_preproc.nii*')))[0]
                noise_std = self._autoload_noise_std(subject_id, file_path=pjoin('noise_std'))

                bval_fname = pjoin('bvals.txt')
                if os.path.isfile(pjoin('diff_preproc.bval')):
                    bval_fname = pjoin('diff_preproc.bval')

                bvec_fname = pjoin('bvecs_fsl_moco_norm.txt')
                if os.path.isfile(pjoin('diff_preproc.bvec')):
                    bvec_fname = pjoin('diff_preproc.bvec')

                prtcl_fname = None
                if os.path.isfile(pjoin('diff_preproc.prtcl')):
                    prtcl_fname = pjoin('diff_preproc.prtcl')

                mask_fname = None
                if list(glob.glob(pjoin('diff_preproc_mask.nii*'))):
                    mask_fname = list(glob.glob(pjoin('diff_preproc_mask.nii*')))[0]

                if mask_fname is None:
                    if list(glob.glob(pjoin('mri', 'diff_preproc_mask.nii*'))):
                        mask_fname = list(glob.glob(pjoin('mri', 'diff_preproc_mask.nii*')))[0]

                protocol_loader = BatchFitProtocolLoader(
                    pjoin(),
                    protocol_fname=prtcl_fname, bvec_fname=bvec_fname, bval_fname=bval_fname,
                    protocol_columns={'Delta': 21.8e-3, 'delta': 12.9e-3, 'TR': 8800e-3, 'TE': 57e-3})

                output_dir = self._get_subject_output_dir(subject_id, mask_fname)

                subjects.append(SimpleSubjectInfo(subject_id, dwi_fname, protocol_loader, mask_fname, output_dir,
                                                  noise_std=noise_std))
        return subjects

    def __str__(self):
        return meta_info['title']



# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2015, 12, 12, 11, 11, 12, 5)

__all__ = [
    'CompositeModelExpressionParser',
    'CompositeModelExpressionSemantics',
]


class CompositeModelExpressionParser(Parser):
    def __init__(self,
                 whitespace=None,
                 nameguard=None,
                 comments_re=None,
                 eol_comments_re=None,
                 ignorecase=None,
                 left_recursion=True,
                 **kwargs):
        super(CompositeModelExpressionParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            **kwargs
        )

    @graken()
    def _result_(self):
        self._expr_()

    @graken()
    def _expr_(self):
        with self._choice():
            with self._option():
                self._term_()
                with self._group():
                    with self._choice():
                        with self._option():
                            self._token('+')
                        with self._option():
                            self._token('-')
                        self._error('expecting one of: + -')
                self._expr_()
            with self._option():
                self._term_()
            self._error('no available options')

    @graken()
    def _term_(self):
        with self._choice():
            with self._option():
                self._factor_()
                with self._group():
                    with self._choice():
                        with self._option():
                            self._token('*')
                        with self._option():
                            self._token('/')
                        self._error('expecting one of: * /')
                self._term_()
            with self._option():
                self._factor_()
            self._error('no available options')

    @graken()
    def _factor_(self):
        with self._choice():
            with self._option():
                self._token('(')
                self._expr_()
                self._token(')')
            with self._option():
                self._model_()
            self._error('no available options')

    @graken()
    def _model_(self):
        self._model_name_()
        with self._optional():
            self._token('(')
            self._nickname_()
            self._token(')')

    @graken()
    def _model_name_(self):
        self._pattern(r'[a-zA-Z_]\w*')

    @graken()
    def _nickname_(self):
        self._pattern(r'[a-zA-Z_]\w*')


class CompositeModelExpressionSemantics(object):
    def result(self, ast):
        return ast

    def expr(self, ast):
        return ast

    def term(self, ast):
        return ast

    def factor(self, ast):
        return ast

    def model(self, ast):
        return ast

    def model_name(self, ast):
        return ast

    def nickname(self, ast):
        return ast



from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco.Lagos'

"""Lineal T1 fitting (Weiskopf, 2016 ESMRMB Workshop)

This fitting is the extension of the standard GRE equation for flip angles lower than 90deg. This modelling allows a
linear fitting of the data if is enough data to support it. In principle, it should not be a problem if only two
points are used, however the addition of a constant in the equation could give some kind of uncertainty.

B1 has to be normalized *in function of the reference voltage, the angle distribution and the reference angle*.
Here I assume that TR <<< T1, then exp(-TR/T1) ~ 1 - TR/T1. Then the equation becomes 'simpler'. However, fi this condition
is not achieved, then return to the standard equation.
Also, DATA HAS TO BE PROCESSED BEFORE TO USE THIS EQUATION. Please apply log() on the data.
"""


class LinT1GRE(CompartmentConfig):

    parameter_list = ('Sw_static', 'E1')
    #cl_code = """
    #    return sin( B1 * angle ) / ( 1 - cos( B1 * angle ) * exp( - TR / T1 ));
    #"""
    cl_code = """
        return Sw_static * E1;
    """


#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Generate a protocol from a bvec and bval file.

MDT uses a protocol file (with extension .prtcl) to store all the acquisition related values.
This is a column based file which can hold, next to the b-values and gradient directions,
the big Delta, small delta, gradient amplitude G and more of these extra acquisition details.
"""
import argparse
import os
from argcomplete.completers import FilesCompleter
import textwrap
import mdt.protocols
from mdt.shell_utils import BasicShellApplication, get_citation_message
from mdt.protocols import load_bvec_bval

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GenerateProtocol(BasicShellApplication):

    def __init__(self):
        mdt.init_user_settings(pass_if_exists=True)

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += get_citation_message()

        epilog = textwrap.dedent("""
        Examples of use:
            mdt-generate-protocol data.bvec data.bval
            mdt-generate-protocol data.bvec data.bval -o my_protocol.prtcl
            mdt-generate-protocol data.bvec data.bval
            mdt-generate-protocol data.bvec data.bval --Delta 30 --delta 20
            mdt-generate-protocol data.bvec data.bval --sequence-timing-units 's' --Delta 0.03
            mdt-generate-protocol data.bvec data.bval --TE ../my_TE_file.txt
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('bvec', help='the gradient vectors file').completer = FilesCompleter()
        parser.add_argument('bval', help='the gradient b-values').completer = FilesCompleter()
        parser.add_argument('-s', '--bval-scale-factor', type=float,
                            help="We expect the b-values in the output protocol in units of s/m^2. "
                                 "Example use: 1 or 1e6. The default is autodetect.")

        parser.add_argument('-o', '--output_file',
                            help='the output protocol, defaults to "<bvec_name>.prtcl" in the same '
                                 'directory as the bvec file.').completer = FilesCompleter()

        parser.add_argument('--sequence-timing-units', choices=('ms', 's'), default='ms',
                            help="The units of the sequence timings. The default is 'ms' which we will convert to 's'.")

        parser.add_argument('--G',
                            help="The gradient amplitudes in T/m.")

        parser.add_argument('--maxG',
                            help="The maximum gradient amplitude in T/m. This is only useful if we need to guess "
                                 "big Delta and small delta. Default is 0.04 T/m")

        parser.add_argument('--Delta',
                            help="The big Delta to use, either a single number or a file with either a single number "
                                 "or one number per gradient direction.")

        parser.add_argument('--delta',
                            help="The small delta to use, either a single number or a file with either a single number "
                                 "or one number per gradient direction.")

        parser.add_argument('--TE',
                            help="The TE to use, either a single number or a file with either a single number "
                                 "or one number per gradient direction.")

        parser.add_argument('--TR',
                            help="The TR to use, either a single number or a file with either a single number "
                                 "or one number per gradient direction.")

        return parser

    def run(self, args):
        bvec = os.path.realpath(args.bvec)
        bval = os.path.realpath(args.bval)

        if args.output_file:
            output_prtcl = os.path.realpath(args.output_file)
        else:
            output_prtcl = os.path.join(os.path.dirname(bvec),
                                        os.path.splitext(os.path.basename(bvec))[0] + '.prtcl')

        if args.bval_scale_factor:
            bval_scale_factor = float(args.bval_scale_factor)
        else:
            bval_scale_factor = 'auto'

        protocol = load_bvec_bval(bvec=bvec, bval=bval, bval_scale=bval_scale_factor)

        if args.G is None and args.maxG is not None:
            if os.path.isfile(str(args.maxG)):
                protocol.add_column_from_file('maxG', os.path.realpath(str(args.maxG)), 1)
            else:
                protocol.add_column('maxG', float(args.maxG))

        if args.Delta is not None:
            add_sequence_timing_column_to_protocol(protocol, 'Delta', args.Delta, args.sequence_timing_units)
        if args.delta is not None:
            add_sequence_timing_column_to_protocol(protocol, 'delta', args.delta, args.sequence_timing_units)
        if args.TE is not None:
            add_sequence_timing_column_to_protocol(protocol, 'TE', args.TE, args.sequence_timing_units)
        if args.TR is not None:
            add_sequence_timing_column_to_protocol(protocol, 'TR', args.TR, args.sequence_timing_units)
        if args.G is not None:
            add_column_to_protocol(protocol, 'G', args.G, 1)

        mdt.protocols.write_protocol(protocol, output_prtcl)


def add_column_to_protocol(protocol, column, value, mult_factor):
    if value is not None:
        if os.path.isfile(value):
            protocol.add_column_from_file(column, os.path.realpath(value), mult_factor)
        else:
            protocol.add_column(column, float(value) * mult_factor)


def add_sequence_timing_column_to_protocol(protocol, column, value, units):
    mult_factor = 1e-3 if units == 'ms' else 1
    add_column_to_protocol(protocol, column, value, mult_factor)


if __name__ == '__main__':
    GenerateProtocol().start()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'save_image_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_SaveImageDialog(object):
    def setupUi(self, SaveImageDialog):
        SaveImageDialog.setObjectName("SaveImageDialog")
        SaveImageDialog.resize(722, 253)
        self.verticalLayout = QtWidgets.QVBoxLayout(SaveImageDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_3 = QtWidgets.QLabel(SaveImageDialog)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_3.addWidget(self.label_3)
        self.label_4 = QtWidgets.QLabel(SaveImageDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_3.addWidget(self.label_4)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.line = QtWidgets.QFrame(SaveImageDialog)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.outputFile_chooser = QtWidgets.QPushButton(SaveImageDialog)
        self.outputFile_chooser.setEnabled(True)
        self.outputFile_chooser.setObjectName("outputFile_chooser")
        self.horizontalLayout.addWidget(self.outputFile_chooser)
        self.outputFile_box = QtWidgets.QLineEdit(SaveImageDialog)
        self.outputFile_box.setEnabled(True)
        self.outputFile_box.setObjectName("outputFile_box")
        self.horizontalLayout.addWidget(self.outputFile_box)
        self.horizontalLayout.setStretch(1, 1)
        self.gridLayout.addLayout(self.horizontalLayout, 3, 1, 1, 1)
        self.label_12 = QtWidgets.QLabel(SaveImageDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 0, 2, 1, 1)
        self.label_11 = QtWidgets.QLabel(SaveImageDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 2, 2, 1, 1)
        self.width_box = QtWidgets.QSpinBox(SaveImageDialog)
        self.width_box.setMinimum(100)
        self.width_box.setMaximum(9999)
        self.width_box.setProperty("value", 1280)
        self.width_box.setObjectName("width_box")
        self.gridLayout.addWidget(self.width_box, 0, 1, 1, 1)
        self.label_14 = QtWidgets.QLabel(SaveImageDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 1, 2, 1, 1)
        self.label_6 = QtWidgets.QLabel(SaveImageDialog)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(SaveImageDialog)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 3, 0, 1, 1)
        self.label = QtWidgets.QLabel(SaveImageDialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)
        self.height_box = QtWidgets.QSpinBox(SaveImageDialog)
        self.height_box.setMinimum(100)
        self.height_box.setMaximum(9999)
        self.height_box.setProperty("value", 720)
        self.height_box.setObjectName("height_box")
        self.gridLayout.addWidget(self.height_box, 1, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(SaveImageDialog)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 2, 0, 1, 1)
        self.dpi_box = QtWidgets.QSpinBox(SaveImageDialog)
        self.dpi_box.setMinimum(10)
        self.dpi_box.setMaximum(1000)
        self.dpi_box.setProperty("value", 100)
        self.dpi_box.setObjectName("dpi_box")
        self.gridLayout.addWidget(self.dpi_box, 2, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.line_3 = QtWidgets.QFrame(SaveImageDialog)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.buttonBox = QtWidgets.QDialogButtonBox(SaveImageDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(SaveImageDialog)
        self.buttonBox.accepted.connect(SaveImageDialog.accept)
        self.buttonBox.rejected.connect(SaveImageDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(SaveImageDialog)
        SaveImageDialog.setTabOrder(self.dpi_box, self.outputFile_chooser)
        SaveImageDialog.setTabOrder(self.outputFile_chooser, self.outputFile_box)

    def retranslateUi(self, SaveImageDialog):
        _translate = QtCore.QCoreApplication.translate
        SaveImageDialog.setWindowTitle(_translate("SaveImageDialog", "Save Image"))
        self.label_3.setText(_translate("SaveImageDialog", "Save image"))
        self.label_4.setText(_translate("SaveImageDialog", "Export the current view to an image."))
        self.outputFile_chooser.setText(_translate("SaveImageDialog", "Browse"))
        self.label_12.setText(_translate("SaveImageDialog", "(The width of the image in pixels)"))
        self.label_11.setText(_translate("SaveImageDialog", "(Dots/pixels per inch)"))
        self.label_14.setText(_translate("SaveImageDialog", "(The height of the image in pixels)"))
        self.label_6.setText(_translate("SaveImageDialog", "Width:"))
        self.label_7.setText(_translate("SaveImageDialog", "Output file:"))
        self.label.setText(_translate("SaveImageDialog", "Height:"))
        self.label_5.setText(_translate("SaveImageDialog", "DPI:"))




from mdt.processing_strategies import ChunksProcessingStrategy

__author__ = 'Robbert Harms'
__date__ = "2015-11-29"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


meta_info = {'title': 'All slices at once',
             'description': 'Processes the whole dataset at once.'}


class AllVoxelsAtOnce(ChunksProcessingStrategy):
    """Run all slices at once."""

    def _chunks_generator(self, model, problem_data, output_path, worker, total_roi_indices):
        yield total_roi_indices



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'fit_model_tab.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_FitModelTabContent(object):
    def setupUi(self, FitModelTabContent):
        FitModelTabContent.setObjectName("FitModelTabContent")
        FitModelTabContent.resize(1047, 427)
        self.verticalLayout = QtWidgets.QVBoxLayout(FitModelTabContent)
        self.verticalLayout.setContentsMargins(-1, 11, -1, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.verticalLayout.addLayout(self.verticalLayout_2)
        self.line = QtWidgets.QFrame(FitModelTabContent)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setHorizontalSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.line_3 = QtWidgets.QFrame(FitModelTabContent)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout.addWidget(self.line_3, 4, 0, 1, 3)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.selectDWI = QtWidgets.QPushButton(FitModelTabContent)
        self.selectDWI.setObjectName("selectDWI")
        self.horizontalLayout_2.addWidget(self.selectDWI)
        self.selectedDWI = QtWidgets.QLineEdit(FitModelTabContent)
        self.selectedDWI.setText("")
        self.selectedDWI.setObjectName("selectedDWI")
        self.horizontalLayout_2.addWidget(self.selectedDWI)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(FitModelTabContent)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.line_4 = QtWidgets.QFrame(FitModelTabContent)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 6, 0, 1, 3)
        self.label_11 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_11.setFont(font)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 2, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(FitModelTabContent)
        self.label_3.setMinimumSize(QtCore.QSize(0, 0))
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 2, 1, 1)
        self.label_12 = QtWidgets.QLabel(FitModelTabContent)
        self.label_12.setMinimumSize(QtCore.QSize(0, 0))
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 5, 0, 1, 1)
        self.modelSelection = QtWidgets.QComboBox(FitModelTabContent)
        self.modelSelection.setObjectName("modelSelection")
        self.gridLayout.addWidget(self.modelSelection, 5, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 1, 2, 1, 1)
        self.label_13 = QtWidgets.QLabel(FitModelTabContent)
        self.label_13.setMinimumSize(QtCore.QSize(0, 0))
        self.label_13.setObjectName("label_13")
        self.gridLayout.addWidget(self.label_13, 3, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(FitModelTabContent)
        self.label_10.setMinimumSize(QtCore.QSize(0, 0))
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 2, 0, 1, 1)
        self.label_14 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 3, 2, 1, 1)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.selectOutputFolder = QtWidgets.QPushButton(FitModelTabContent)
        self.selectOutputFolder.setObjectName("selectOutputFolder")
        self.horizontalLayout_7.addWidget(self.selectOutputFolder)
        self.selectedOutputFolder = QtWidgets.QLineEdit(FitModelTabContent)
        self.selectedOutputFolder.setText("")
        self.selectedOutputFolder.setObjectName("selectedOutputFolder")
        self.horizontalLayout_7.addWidget(self.selectedOutputFolder)
        self.gridLayout.addLayout(self.horizontalLayout_7, 3, 1, 1, 1)
        self.label_15 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_15.setFont(font)
        self.label_15.setObjectName("label_15")
        self.gridLayout.addWidget(self.label_15, 5, 2, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.selectMask = QtWidgets.QPushButton(FitModelTabContent)
        self.selectMask.setObjectName("selectMask")
        self.horizontalLayout_3.addWidget(self.selectMask)
        self.selectedMask = QtWidgets.QLineEdit(FitModelTabContent)
        self.selectedMask.setText("")
        self.selectedMask.setObjectName("selectedMask")
        self.horizontalLayout_3.addWidget(self.selectedMask)
        self.gridLayout.addLayout(self.horizontalLayout_3, 1, 1, 1, 1)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.selectProtocol = QtWidgets.QPushButton(FitModelTabContent)
        self.selectProtocol.setObjectName("selectProtocol")
        self.horizontalLayout_6.addWidget(self.selectProtocol)
        self.selectedProtocol = QtWidgets.QLineEdit(FitModelTabContent)
        self.selectedProtocol.setText("")
        self.selectedProtocol.setObjectName("selectedProtocol")
        self.horizontalLayout_6.addWidget(self.selectedProtocol)
        self.gridLayout.addLayout(self.horizontalLayout_6, 2, 1, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.optimizationOptionsButton = QtWidgets.QPushButton(FitModelTabContent)
        self.optimizationOptionsButton.setObjectName("optimizationOptionsButton")
        self.horizontalLayout_4.addWidget(self.optimizationOptionsButton)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_4, 7, 1, 1, 1)
        self.label_17 = QtWidgets.QLabel(FitModelTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_17.setFont(font)
        self.label_17.setObjectName("label_17")
        self.gridLayout.addWidget(self.label_17, 7, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line_2 = QtWidgets.QFrame(FitModelTabContent)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 6, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.runButton = QtWidgets.QPushButton(FitModelTabContent)
        self.runButton.setEnabled(True)
        self.runButton.setObjectName("runButton")
        self.horizontalLayout.addWidget(self.runButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)
        self.line_2.raise_()
        self.line.raise_()

        self.retranslateUi(FitModelTabContent)
        QtCore.QMetaObject.connectSlotsByName(FitModelTabContent)
        FitModelTabContent.setTabOrder(self.selectDWI, self.selectedDWI)
        FitModelTabContent.setTabOrder(self.selectedDWI, self.selectMask)
        FitModelTabContent.setTabOrder(self.selectMask, self.selectedMask)
        FitModelTabContent.setTabOrder(self.selectedMask, self.selectProtocol)
        FitModelTabContent.setTabOrder(self.selectProtocol, self.selectedProtocol)
        FitModelTabContent.setTabOrder(self.selectedProtocol, self.selectOutputFolder)
        FitModelTabContent.setTabOrder(self.selectOutputFolder, self.selectedOutputFolder)
        FitModelTabContent.setTabOrder(self.selectedOutputFolder, self.modelSelection)
        FitModelTabContent.setTabOrder(self.modelSelection, self.runButton)

    def retranslateUi(self, FitModelTabContent):
        _translate = QtCore.QCoreApplication.translate
        FitModelTabContent.setWindowTitle(_translate("FitModelTabContent", "Form"))
        self.label.setText(_translate("FitModelTabContent", "Fit model"))
        self.label_2.setText(_translate("FitModelTabContent", "Optimize a model to your data."))
        self.selectDWI.setText(_translate("FitModelTabContent", "Browse"))
        self.label_6.setText(_translate("FitModelTabContent", "Select DWI:"))
        self.label_11.setText(_translate("FitModelTabContent", "(Select your protocol file, see tab \"Generate protocol file\")"))
        self.label_3.setText(_translate("FitModelTabContent", "Select brain mask:"))
        self.label_4.setText(_translate("FitModelTabContent", "(Select your preprocessed 4d diffusion weighted volume)"))
        self.label_12.setText(_translate("FitModelTabContent", "Select model:"))
        self.label_5.setText(_translate("FitModelTabContent", "(Select your brain mask, see tab \"Generate brain mask\")"))
        self.label_13.setText(_translate("FitModelTabContent", "Select output folder:"))
        self.label_10.setText(_translate("FitModelTabContent", "Select protocol file:"))
        self.label_14.setText(_translate("FitModelTabContent", "(Defaults to \"output/<mask_name>\" in the DWI directory)"))
        self.selectOutputFolder.setText(_translate("FitModelTabContent", "Browse"))
        self.label_15.setText(_translate("FitModelTabContent", "(Please select a model)"))
        self.selectMask.setText(_translate("FitModelTabContent", "Browse"))
        self.selectProtocol.setText(_translate("FitModelTabContent", "Browse"))
        self.optimizationOptionsButton.setText(_translate("FitModelTabContent", "Optimization options"))
        self.label_17.setText(_translate("FitModelTabContent", "(Additional settings)"))
        self.runButton.setText(_translate("FitModelTabContent", "Run"))




import glob
import os
import numpy as np
import nibabel as nib
from mdt.deferred_mappings import DeferredActionDict

__author__ = 'Robbert Harms'
__date__ = "2014-08-28"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def load_nifti(nifti_volume):
    """Load and return a nifti file.

    This will apply path resolution if a filename without extension is given. See the function
    :func:`nifti_filepath_resolution` for details.

    Args:
        nifti_volume (string): The filename of the volume to use.

    Returns:
        :class:`nibabel.nifti1.Nifti1Image`
    """
    path = nifti_filepath_resolution(nifti_volume)
    return nib.load(path)


def load_all_niftis(directory, map_names=None):
    """Loads all niftis in the given directory.

    If map_names is given we will only load the given maps. Else, we load all .nii and .nii.gz files in the
    given directory. The map name is the filename of a nifti without the extension.

    Args:
        directory (str): the directory from which we want to load the niftis
        map_names (list of str): the names of the maps we want to use. If given, we only use and return these maps.

    Returns:
        dict: A dictionary with the loaded nibabel proxies (see :func:`load_nifti`).
            The keys of the dictionary are the filenames without the extension of the .nii(.gz) files
            in the given directory.
    """
    maps_paths = {}

    for path, map_name, _ in yield_nifti_info(directory):
        if not map_names or map_name in map_names:
            maps_paths.update({map_name: path})

    return {k: load_nifti(v) for k, v in maps_paths.items()}


def get_all_image_data(directory, map_names=None, deferred=True):
    """Get the data of all the nifti volumes in the given directory.

    If map_names is given we will only load the given map names. Else, we load all .nii and .nii.gz files in the
    given directory.

    Args:
        directory (str): the directory from which we want to read a number of maps
        map_names (list of str): the names of the maps we want to use. If given, we only use and return these maps.
        deferred (boolean): if True we return an deferred loading dictionary instead of a dictionary with the values
            loaded as arrays.

    Returns:
        dict: A dictionary with the volumes. The keys of the dictionary are the filenames
            without the extension of the .nii(.gz) files in the given directory.
    """
    proxies = load_all_niftis(directory, map_names=map_names)
    if deferred:
        return DeferredActionDict(lambda _, item: item.get_data(), proxies)
    else:
        return {k: v.get_data() for k, v in proxies.items()}


def write_nifti(data, header, output_fname, affine=None, **kwargs):
    """Write data to a nifti file.

    Args:
        output_fname (str): the name of the resulting nifti file
        data (ndarray): the data to write to that nifti file
        header (nibabel header): the nibabel header to use as header for the nifti file
        affine (ndarray): the affine transformation matrix
        **kwargs: other arguments to Nifti1Image from NiBabel

    """
    nib.Nifti1Image(data, affine, header, **kwargs).to_filename(output_fname)


def write_all_as_nifti(volumes, directory, nifti_header, overwrite_volumes=True, gzip=True):
    """Write a number of volume maps to the specific directory.

    Args:
        volumes (dict): the volume maps (in 3d) with the results we want to write.
            The filenames are generated using the keys in the given volumes
        directory (str): the directory to write to
        nifti_header: the nifti header to use for each of the volumes
        overwrite_volumes (boolean): defaults to True, if we want to overwrite the volumes if they exists
        gzip (boolean): if True we write the files as .nii.gz, if False we write the files as .nii
    """
    if not os.path.exists(directory):
        os.makedirs(directory)

    for key, volume in volumes.items():
        extension = '.nii'
        if gzip:
            extension += '.gz'
        filename = key + extension

        full_filename = os.path.abspath(os.path.join(directory, filename))

        if os.path.exists(full_filename):
            if overwrite_volumes:
                os.remove(full_filename)
                write_nifti(volume, nifti_header, full_filename)
        else:
            write_nifti(volume, nifti_header, full_filename)


def nifti_filepath_resolution(file_path):
    """Tries to resolve the filename to a nifti based on only the filename.

    For example, this resolves the path: ``/tmp/mask`` to:

        - ``/tmp/mask`` if exists
        - ``/tmp/mask.nii`` if exist
        - ``/tmp/mask.nii.gz`` if exists

    Hence, the lookup order is: ``path``, ``path.nii``, ``path.nii.gz``

    If a file with an extension is given we will do no further resolving and return the path as is.

    Args:
        file_path (str): the path to the nifti file, can be without extension.

    Returns:
        str: the file path we resolved to the final file.

    Raises:
        ValueError: if no nifti file could be found
    """
    if file_path[:-len('.nii')] == '.nii' or file_path[:-len('.nii.gz')] == '.nii.gz':
        return file_path

    if os.path.isfile(file_path):
        return file_path
    elif os.path.isfile(file_path + '.nii'):
        return file_path + '.nii'
    elif os.path.isfile(file_path + '.nii.gz'):
        return file_path + '.nii.gz'
    raise ValueError('No nifti file could be found using the path {}.'.format(file_path))


def yield_nifti_info(directory):
    """Get information about the nifti volumes in the given directory.

    Args:
        directory (str): the directory to get the names of the available maps from

    Yields:
        tuple: (path, map_name, extension) for every map found
    """
    for extension in ('.nii', '.nii.gz'):
        for f in glob.glob(os.path.join(directory, '*' + extension)):
            yield f, os.path.basename(f)[0:-len(extension)], extension



class TrackMark(object):
    """TrackMark is an proprietary visualization tool written by Alard Roebroeck and can be used to visualize fibre\
    directions. This class is meant to convert nifti files to TrackMark specific files.
    """

    @staticmethod
    def write_tvl_direction_pairs(tvl_filename, tvl_header, direction_pairs):
        """Write the given directions to TVL.

        The direction pairs should be a list with lists containing the vector and value to write. For example:
        ((vec, val), (vec1, val1), ...) up to three pairs are allowed.

        Args:
            tvl_filename (str): the filename to write to
            tvl_header (:class:`list`): the header for the TVL file. This is a list of either 4 or 10 entries.
                4 entries: [version, res, gap, offset]
                10 entries: [version, x_res, x_gap, x_offset, y_res, y_gap, y_offset, z_res, z_gap, z_offset]
            direction_pairs (list of ndarrays): The list with direction pairs, only three are used.
                This is a list with (vector, magnitude) tuples in which the vectors are 4d volumes with for
                every voxel a 3d coordinate.
        """
        direction_pairs = direction_pairs[0:3]
        dir_matrix = np.zeros(direction_pairs[0][0].shape[0:3] + (12,))
        for ind, dirs in enumerate(direction_pairs):
            dir_matrix[..., ind*3:ind*3+3] = np.ascontiguousarray(TrackMark._ensure_3d(np.squeeze(dirs[0])))
            dir_matrix[..., 9 + ind] = np.ascontiguousarray(TrackMark._ensure_3d(np.squeeze(dirs[1])))

        TrackMark.write_tvl_matrix(tvl_filename, tvl_header, dir_matrix)

    @staticmethod
    def write_tvl_matrix(tvl_filename, tvl_header, directions_matrix):
        """Write the given directions matrix to TVL.

        Args:
            tvl_filename: the filename to write to
            tvl_header: the header for the TVL file. This is a list of either 4 or 10 entries.
                4 entries: [version, res, gap, offset]
                10 entries: [version, x_res, x_gap, x_offset, y_res, y_gap, y_offset, z_res, z_gap, z_offset]
            directions_matrix: an 4dimensional matrix, of which the fourth dimension is of length 12.
        """
        if os.path.exists(tvl_filename):
            os.remove(tvl_filename)

        if not os.path.exists(os.path.dirname(tvl_filename)):
            os.makedirs(os.path.dirname(tvl_filename))

        open(tvl_filename, 'a').close()
        with open(tvl_filename, 'rb+') as f:
            version = np.array(tvl_header[0]).astype(np.uint16)
            version.tofile(f, '')

            if len(tvl_header) == 4:
                for i in range(3):
                    np.array(directions_matrix.shape[i]).astype(np.uint32).tofile(f, '')
                    np.array(tvl_header[1]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[2]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[3]).astype(np.float64).tofile(f, '')
            else:
                for i in range(3):
                    np.array(directions_matrix.shape[i]).astype(np.uint32).tofile(f, '')
                    np.array(tvl_header[i * 3 + 1]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[i * 3 + 2]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[i * 3 + 3]).astype(np.float64).tofile(f, '')

            directions_matrix = np.transpose(directions_matrix, (3, 2, 1, 0)).astype(np.float32).flatten('F')
            directions_matrix.tofile(f, '')

    @staticmethod
    def write_rawmaps(directory, volumes, overwrite_volumes=True):
        """Write a dictionary with volumes to the given directory.

        Args:
            directory (str): the directory to write to
            volumes (dict): an dictionary with the volume maps (3d) with the results we want to write out
                The naming of the file is the key of the volume with .rawmap appended by this function.
            overwrite_volumes (boolean): if we want to overwrite already present volumes
        """
        if not os.path.isdir(directory):
            os.makedirs(directory)

        for key, volume in volumes.items():
            filename = key + '.rawmap'
            full_filename = os.path.abspath(os.path.join(directory, filename))

            if os.path.exists(full_filename):
                if overwrite_volumes:
                    os.remove(full_filename)
                    TrackMark.write_rawmap(full_filename, volume)
            else:
                TrackMark.write_rawmap(full_filename, volume)

    @staticmethod
    def write_rawmap(rawmap_filename, volume):
        """Write a rawmap to the given file.

        Args:
            rawmap_filename (str): The filename to write to, if not exists, it is created
                (along with extra directories). This should end on .rawmap, it not it is added.
            volume (ndarray): the volume to write. 3d or 4d. If 4d and 4th dimension is larger than 1
                additional maps are created.
        """
        if rawmap_filename[-len('.rawmap'):] != '.rawmap':
            rawmap_filename += '.rawmap'

        if os.path.exists(rawmap_filename):
            os.remove(rawmap_filename)

        if not os.path.exists(os.path.dirname(rawmap_filename)):
            os.makedirs(os.path.dirname(rawmap_filename))

        open(rawmap_filename, 'a').close()

        s = volume.shape
        if len(s) == 4:
            if s[3] == 1:
                volume = np.squeeze(volume, axis=(3,))
            else:
                subnames = rawmap_filename[0:-len('.rawmap')] + '_'
                for ind in range(volume.shape[3]):
                    TrackMark.write_rawmap(subnames + repr(ind) + '.rawmap', volume[..., ind])
                return

        with open(rawmap_filename, 'rb+') as f:
            np.array(s[0]).astype(np.uint16).tofile(f, '')
            np.array(s[1]).astype(np.uint16).tofile(f, '')
            np.array(s[2]).astype(np.uint16).tofile(f, '')

            m = np.transpose(volume, [2, 1, 0]).astype(np.float32).flatten('F')
            m.tofile(f, '')

    @staticmethod
    def auto_convert(input_folder, model_name=None, output_folder=None):
        """Convert the nifti files in the given folder to Trackmark.

        This automatically loads the correct files based on the model name. This is normally the dirname of the given
        path. If that is not the case you can give the model name explicitly.

        By default it outputs the results to a folder named "trackmark" in the given input folder. This can of course
        be overridden using the output_folder parameter.

        Args:
            input_folder (str): the name of the input folder
            model_name (str): the name of the model, if not given we use the last dirname of the given path
            output_folder (str): the output folder, if not given we will output to a subfolder "trackmark" in the
                given directory.
        """
        output_folder = output_folder or os.path.join(input_folder, 'trackmark')
        model_name = model_name or os.path.basename(os.path.normpath(input_folder))

        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

        conversion_profile = get_trackmark_conversion_profile(model_name)
        direction_pairs, volumes = conversion_profile.get_info(input_folder)

        tvl_header = (1, 1.8, 0, 0)

        TrackMark.write_rawmaps(output_folder, volumes)
        TrackMark.write_tvl_direction_pairs(output_folder + '/master.tvl', tvl_header, direction_pairs)

    @staticmethod
    def _ensure_3d(array):
        if len(array.shape) < 3:
            return array[..., None]
        return array


def get_trackmark_conversion_profile(model_name):
    """Get the TrackMark conversion profile for the given model name.

    Args:
        model_name (str): the name of the model

    Returns:
        TrackMarkConversionProfile: the trackmark conversion profile for the given model name

    Raises:
        ValueError: if no conversion profile for the given model name could be found.
    """
    if '_TMCP_' + model_name.replace('-ExVivo', '') in globals():
        return globals()['_TMCP_' + model_name.replace('-ExVivo', '')]()

    raise ValueError('No trackmark conversion profile could be found for the model named {}.'.format(model_name))


class TrackMarkConversionProfile(object):

    def get_info(self, input_folder):
        """Create the trackmark files (TVL and rawmap) using this profile for the data in the given input folder

        Args:
            input_folder (str): the folder containing the niftis to convert

        Returns:
            tuple: the (direction_paris, volumes) tuple. The first should contain the direction pairs
                needed for the TVL output, the second is the list of volumes to write as rawmaps.
        """


class _TMCP_BallStick_r1(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        maps_to_convert = ['FS', 'w_ball.w', 'w_stick.w']

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)

        vector_directions = [load_nifti(os.path.join(input_folder, 'Stick.vec0')).get_data()]
        vector_magnitudes = [load_nifti(os.path.join(input_folder, 'w_stick.w')).get_data() * 1e-2]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_BallStick_r2(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        import mdt

        maps_to_convert = ['FS', 'w_ball.w']

        sort_output = mdt.sort_maps(
            [os.path.join(input_folder, 'w_stick{}.w.nii.gz'.format(i)) for i in range(2)],
            extra_maps_to_sort=[os.path.join(input_folder, 'Stick{}.vec0.nii.gz'.format(i)) for i in range(2)],
            reversed_sort=True)

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)
        volumes.update(dict(zip(['w_stick{}.w'.format(i) for i in range(2)], sort_output[0])))

        vector_directions = sort_output[1]
        vector_magnitudes = [v * 1e-2 for v in sort_output[0]]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_BallStick_r3(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        import mdt

        maps_to_convert = ['FS', 'w_ball.w']

        sort_output = mdt.sort_maps(
            [os.path.join(input_folder, 'w_stick{}.w.nii.gz'.format(i)) for i in range(3)],
            extra_maps_to_sort=[os.path.join(input_folder, 'Stick{}.vec0.nii.gz'.format(i)) for i in range(3)],
            reversed_sort=True)

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)
        volumes.update(dict(zip(['w_stick{}.w'.format(i) for i in range(3)], sort_output[0])))

        vector_directions = sort_output[1]
        vector_magnitudes = [v * 1e-2 for v in sort_output[0]]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_CHARMED_r1(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        maps_to_convert = ['FR', 'Tensor.FA', 'w_hin0.w']

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)

        vector_directions = [load_nifti(os.path.join(input_folder, 'CHARMEDRestricted0.vec0')).get_data()]
        vector_magnitudes = [load_nifti(os.path.join(input_folder, 'w_res0.w')).get_data() * 1e-2]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_CHARMED_r2(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        import mdt

        maps_to_convert = ['FR', 'Tensor.FA', 'w_hin0.w']

        sort_output = mdt.sort_maps(
            [os.path.join(input_folder, 'w_res{}.w.nii.gz'.format(i)) for i in range(2)],
            extra_maps_to_sort=[os.path.join(input_folder, 'CHARMEDRestricted{}.vec0.nii.gz'.format(i))
                                for i in range(2)],
            reversed_sort=True)

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)
        volumes.update(dict(zip(['CHARMEDRestricted{}.w'.format(i) for i in range(2)], sort_output[0])))

        vector_directions = sort_output[1]
        vector_magnitudes = [v * 1e-2 for v in sort_output[0]]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_CHARMED_r3(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        import mdt

        maps_to_convert = ['FR', 'Tensor.FA', 'w_hin0.w']

        sort_output = mdt.sort_maps(
            [os.path.join(input_folder, 'w_res{}.w.nii.gz'.format(i)) for i in range(3)],
            extra_maps_to_sort=[os.path.join(input_folder, 'CHARMEDRestricted{}.vec0.nii.gz'.format(i))
                                for i in range(3)],
            reversed_sort=True)

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)
        volumes.update(dict(zip(['CHARMEDRestricted{}.w'.format(i) for i in range(3)], sort_output[0])))

        vector_directions = sort_output[1]
        vector_magnitudes = [v * 1e-2 for v in sort_output[0]]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_NODDI(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        maps_to_convert = ['NDI', 'NODDI_EC.kappa', 'ODI', 'w_csf.w', 'w_ec.w', 'w_ic.w']

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)

        vector_directions = [load_nifti(os.path.join(input_folder, 'NODDI_IC.vec0')).get_data()]
        vector_magnitudes = [load_nifti(os.path.join(input_folder, 'w_ic.w')).get_data() * 1e-2]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes


class _TMCP_Tensor(TrackMarkConversionProfile):

    def get_info(self, input_folder):
        maps_to_convert = ['Tensor.FA', 'Tensor.AD', 'Tensor.MD', 'Tensor.RD',
                           'Tensor.sorted_eigval0', 'Tensor.sorted_eigval1', 'Tensor.sorted_eigval2']

        volumes = get_all_image_data(input_folder, map_names=maps_to_convert, deferred=True)

        vector_directions = [load_nifti(os.path.join(input_folder, 'Tensor.sorted_vec{}'.format(i))).get_data()
                             for i in range(3)]
        vector_magnitudes = [load_nifti(os.path.join(input_folder, 'Tensor.sorted_eigval{}'.format(i))).get_data() * 1e6
                             for i in range(3)]

        direction_pairs = list(zip(vector_directions, vector_magnitudes))[:3]

        return direction_pairs, volumes



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Fits a batch profile to a set of data.

This script can be used to fit multiple models to multiple datasets. It needs a batch profile with information
about the subjects. If no batch profile is given, this routine will try to auto-detect a good batch profile.

The most general batch profile is the 'DirPerSubject' profile which assumes that every subject has
its own subdirectory under the given data folder. For details, please look up the batch profiles in your home folder.

A few of the batch profile settings can be altered with arguments to this script. For example,
use_gradient_deviations and models_to_fit override the values in the batch profile.
"""
import argparse
import os
import mdt
from argcomplete.completers import FilesCompleter
from mdt.batch_utils import batch_profile_factory, SelectedSubjects
from mdt.components_loader import BatchProfilesLoader

from mdt.shell_utils import BasicShellApplication
from mot import cl_environments
import textwrap

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BatchFit(BasicShellApplication):

    def __init__(self):
        self.available_devices = list((ind for ind, env in
                                       enumerate(cl_environments.CLEnvironmentFactory.smart_device_selection())))

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += mdt.shell_utils.get_citation_message()

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-batch-fit .
                mdt-batch-fit /data/mgh --batch-profile 'HCP_MGH'
                mdt-batch-fit . --subjects-index 0 1 2 --subjects-id 1003 1004
                mdt-batch-fit . --dry-run
        """)
        batch_profiles = BatchProfilesLoader().list_all()

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('data_folder', help='the directory with the subject to fit').completer = FilesCompleter()

        parser.add_argument('-b', '--batch_profile', default=None, choices=batch_profiles,
                            help='The batch profile (by name) to use during fitting. If not given a'
                                 'batch profile is auto-detected.')

        parser.add_argument('--cl-device-ind', type=int, nargs='*', choices=self.available_devices,
                            help="The index of the device we would like to use. This follows the indices "
                                 "in mdt-list-devices and defaults to the first GPU.")

        parser.add_argument('--recalculate', dest='recalculate', action='store_true',
                            help="Recalculate the model(s) if the output exists.")
        parser.add_argument('--no-recalculate', dest='recalculate', action='store_false',
                            help="Do not recalculate the model(s) if the output exists. (default)")
        parser.set_defaults(recalculate=False)

        parser.add_argument('--use-gradient-deviations', dest='use_gradient_deviations', action='store_true',
                            help="Uses the gradient deviations. If not set, the default in the profile is used.")
        parser.add_argument('--no-gradient-deviations', dest='use_gradient_deviations', action='store_false',
                            help="Disable the use of gradient deviations. If not set, the default "
                                 "in the profile is used.")
        parser.set_defaults(use_gradient_deviations=None)

        parser.add_argument('--double', dest='double_precision', action='store_true',
                            help="Calculate in double precision.")
        parser.add_argument('--float', dest='double_precision', action='store_false',
                            help="Calculate in single precision. (default)")
        parser.set_defaults(double_precision=False)

        parser.add_argument('--subjects-index', type=int, nargs='*',
                            help="The index of the subjects we would like to fit. This reduces the set of"
                                 "subjects.")

        parser.add_argument('--subjects-id', type=str, nargs='*',
                            help="The id of the subjects we would like to fit. This reduces the set of"
                                 "subjects.")

        parser.add_argument('--models-to-fit', type=str, nargs='*',
                            help="The models to fit, this overrides the models in the batch profile.")

        parser.add_argument('--dry-run', dest='dry_run', action='store_true',
                            help="Shows what it will do without the dry run argument.")
        parser.set_defaults(dry_run=False)

        parser.add_argument('--use-cascade-subdir', dest='cascade_subdir', action='store_true',
                            help="Set if you want to create a subdirectory for the given cascade model"
                                 ", default is False.")
        parser.set_defaults(cascade_subdir=False)

        parser.add_argument('--tmp-results-dir', dest='tmp_results_dir', default='True', type=str,
                            help='The directory for the temporary results. The default ("True") uses the config file '
                                 'setting. Set to the literal "None" to disable.').completer = FilesCompleter()

        return parser

    def run(self, args):
        batch_profile = batch_profile_factory(args.batch_profile, os.path.realpath(args.data_folder))

        if args.use_gradient_deviations is not None:
            batch_profile.use_gradient_deviations = args.use_gradient_deviations

        if args.models_to_fit is not None:
            batch_profile.models_to_fit = args.models_to_fit

        subjects_selection = None
        if args.subjects_index or args.subjects_id:
            indices = args.subjects_index if args.subjects_index else []
            subject_ids = args.subjects_id if args.subjects_id else []

            subjects_selection = SelectedSubjects(indices=indices, subject_ids=subject_ids)

        tmp_results_dir = args.tmp_results_dir
        for match, to_set in [('true', True), ('false', False), ('none', None)]:
            if tmp_results_dir.lower() == match:
                tmp_results_dir = to_set
                break

        mdt.batch_fit(os.path.realpath(args.data_folder),
                      subjects_selection=subjects_selection,
                      batch_profile=batch_profile,
                      recalculate=args.recalculate,
                      cl_device_ind=args.cl_device_ind,
                      double_precision=args.double_precision,
                      dry_run=args.dry_run,
                      cascade_subdir=args.cascade_subdir,
                      tmp_results_dir=tmp_results_dir)


if __name__ == '__main__':
    BatchFit().start()



import argparse
import os
import textwrap
import argcomplete
import sys

__author__ = 'Robbert Harms'
__date__ = "2015-10-16"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def get_argparse_extension_checker(choices, dir_allowed=False):
    """Get an :class:`argparge.Action` class that can check for correct extensions.

    Returns:
        argparse.Action: a class (not an instance) of an argparse action.
    """

    class Act(argparse.Action):
        def __call__(self, parser, namespace, fname, option_string=None):
            is_valid = any(map(lambda choice: fname[-len(choice):] == choice, choices))

            if not is_valid and dir_allowed and os.path.isdir(fname):
                is_valid = True

            if is_valid:
                setattr(namespace, self.dest, fname)
            else:
                option_string = '({})'.format(option_string) if option_string else ''
                parser.error("File doesn't end with one of {}{}".format(choices, option_string))

    return Act


class BasicShellApplication(object):

    @classmethod
    def console_script(cls):
        """Method used to start the command when launched from a distutils console script."""
        cls().start(sys.argv[1:])

    def start(self, run_args=None):
        """ Starts a command and registers single handlers.

        Args:
            run_args (:class:`list`): the list of run arguments. If None we use sys.argv[1:].
        """
        if run_args is None:
            run_args = sys.argv[1:]

        parser = self._get_arg_parser()
        argcomplete.autocomplete(parser)
        args = parser.parse_args(run_args)
        self.run(args)

    def run(self, args):
        """Run the application with the given arguments.

        Args:
            args: the arguments from the argparser.
        """

    def _get_arg_parser(self):
        """Create the auto parser. This should be implemented by the implementing class.

        To enable autocomplete in your shell please execute activate-global-python-argcomplete in your shell.

        """
        description = textwrap.dedent("""
            Basic parser introduction here.

            Can be multiline.
        """)

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-model-fit "BallStick (Cascade)" data.nii.gz data.prtcl roi_mask_0_50.nii.gz
        """)
        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)
        return parser

    def _get_citation_message(self):
        return get_citation_message()


def get_citation_message():
    """The citation message used in the shell scripts.

    Returns:
        str: the citation message for use in the description of every shell script
    """
    return textwrap.dedent("""
        If you use any of the scripts/functions/tools from MDT in your research, please cite the following paper:
            <citation here>
    """)





import numpy as np

from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick_r1(CascadeConfig):

    name = 'BallStick_r1 (Cascade)'
    description = 'Cascade for Ballstick'
    models = ('S0',
              'BallStick_r1')

    lower_bounds = {'BallStick_r1': [
        ('S0.s0', lambda output_previous, output_all_previous: 2 * np.min(output_previous['S0.s0']))
    ]}

    upper_bounds = {'BallStick_r1': [
        ('S0.s0', lambda output_previous, output_all_previous: 2 * np.max(output_previous['S0.s0']))
    ]}


class BallStick_r1_S0(BallStick_r1):

    name = 'BallStick_r1 (Cascade|S0)'


class BallStick_r1_ExVivo(BallStick_r1):

    name = 'BallStick_r1-ExVivo (Cascade)'
    description = 'Cascade for Ballstick with ex vivo defaults.'
    models = ('S0',
              'BallStick_r1-ExVivo')


class BallStick_r2(CascadeConfig):

    name = 'BallStick_r2 (Cascade)'
    description = 'Cascade for BallStick_r2.'
    models = ('BallStick_r1 (Cascade)',
              'BallStick_r2')
    inits = {'BallStick_r2': [('Stick0.theta', 'Stick.theta'),
                              ('Stick0.phi', 'Stick.phi'),
                              ('w_stick0.w', 'w_stick.w'),
                              ('w_stick1.w', 0.0)]}


class BallStick_r2_ExVivo(BallStick_r2):

    name = 'BallStick_r2-ExVivo (Cascade)'
    description = 'Cascade for BallStick_r2 with ex vivo defaults.'
    models = ('BallStick_r1-ExVivo (Cascade)',
              'BallStick_r2-ExVivo')


class BallStick_r3(CascadeConfig):

    name = 'BallStick_r3 (Cascade)'
    description = 'Cascade for BallStick_r3.'
    models = ('BallStick_r2 (Cascade)',
              'BallStick_r3')
    inits = {'BallStick_r3': [('w_stick2.w', 0.0)]}


class BallStick_r3_ExVivo(BallStick_r3):

    name = 'BallStick_r3-ExVivo (Cascade)'
    description = 'Cascade for BallStick_r3 with ex vivo defaults.'
    models = ('BallStick_r2-ExVivo (Cascade)',
              'BallStick_r3-ExVivo')



import numpy as np
from mdt import create_roi
from mdt.utils import ComplexNoiseStdEstimator
from mdt.exceptions import NoiseStdEstimationNotPossible

__author__ = 'Robbert Harms'
__date__ = "2015-11-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AllUnweightedVolumes(ComplexNoiseStdEstimator):

    def estimate(self, problem_data, **kwargs):
        """Calculate the standard deviation of the error using all unweighted volumes.

        This calculates per voxel (in the brain mas) the std over all unweighted volumes
        and takes the mean of those estimates as the standard deviation of the noise.

        The method is taken from Camino (http://camino.cs.ucl.ac.uk/index.php?n=Man.Estimatesnr).
        """
        unweighted_indices = problem_data.protocol.get_unweighted_indices()
        unweighted_volumes = problem_data.dwi_volume[..., unweighted_indices]

        if len(unweighted_indices) < 2:
            raise NoiseStdEstimationNotPossible('Not enough unweighted volumes for this estimator.')

        voxel_list = create_roi(unweighted_volumes, problem_data.mask)
        return np.mean(np.std(voxel_list, axis=1))

    def __str__(self):
        return __name__



from pkg_resources import resource_filename
from mot.model_building.cl_functions.base import LibraryFunction

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MRIConstants(LibraryFunction):

    def __init__(self):
        super(MRIConstants, self).__init__(
            '',
            '',
            (),
            resource_filename(__name__, 'MRIConstants.h'),
            resource_filename(__name__, 'MRIConstants.cl'),
            {},
            ())



from contextlib import contextmanager
import logging
import os
import timeit
import time
from mdt.utils import model_output_exists, load_samples
from mdt.processing_strategies import SimpleModelProcessingWorkerGenerator, SamplingProcessingWorker
from mdt.exceptions import InsufficientProtocolError

__author__ = 'Robbert Harms'
__date__ = "2015-05-01"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def sample_composite_model(model, problem_data, output_folder, sampler, processing_strategy,
                        recalculate=False, store_samples=True):
    """Sample a composite model.

    Args:
        model (:class:`~mdt.models.composite.DMRICompositeModel`): a composite model to sample
        problem_data (:class:`~mdt.utils.DMRIProblemData`): The problem data object with which the model
            is initialized before running
        output_folder (string): The full path to the folder where to place the output
        sampler (:class:`mot.cl_routines.sampling.base.AbstractSampler`): The sampling routine to use.
        processing_strategy (:class:`~mdt.processing_strategies.ModelProcessingStrategy`): the processing strategy
        recalculate (boolean): If we want to recalculate the results if they are already present.
        store_samples (boolean): if set to False we will store none of the samples. Use this
            if you are only interested in the volume maps and not in the entire sample chain.
    """
    if not model.is_protocol_sufficient(problem_data.protocol):
        raise InsufficientProtocolError(
            'The provided protocol is insufficient for this model. '
            'The reported errors where: {}'.format(model.get_protocol_problems(problem_data.protocol)))

    logger = logging.getLogger(__name__)

    if not recalculate:
        if model_output_exists(model, output_folder + '/volume_maps/', append_model_name_to_path=False):
            logger.info('Not recalculating {} model'.format(model.name))
            return load_samples(output_folder)

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    model.set_problem_data(problem_data)

    with _log_info(logger, model.name):
        worker_generator = SimpleModelProcessingWorkerGenerator(
            lambda *args: SamplingProcessingWorker(sampler, store_samples, *args))
        return processing_strategy.run(model, problem_data, output_folder, recalculate, worker_generator)


@contextmanager
def _log_info(logger, model_name):
    minimize_start_time = timeit.default_timer()
    logger.info('Sampling {} model'.format(model_name))
    yield
    run_time = timeit.default_timer() - minimize_start_time
    run_time_str = time.strftime('%H:%M:%S', time.gmtime(run_time))
    logger.info('Sampled {0} model with runtime {1} (h:m:s).'.format(model_name, run_time_str))



import collections
import glob
import numbers
import os
import numpy as np
import copy
import six
from mdt.exceptions import ProtocolIOError

__author__ = 'Robbert Harms'
__date__ = "2014-02-06"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Protocol(collections.MutableMapping):

    def __init__(self, columns=None):
        """Create a new protocol. Optionally initializes the protocol with the given set of columns.

        Please note that we use SI units throughout MDT. Take care when loading the data that you load it in SI units.

        For example:

        * G (gradient amplitude) in T/m (Tesla per meter)
        * Delta (time interval) in seconds
        * delta (duration) in seconds


        Args:
            columns (dict): The initial list of columns used by this protocol, the keys should be the name of the
                parameter (the same as those used in the model functions).
                The values should be numpy arrays of equal length.
        """
        super(Protocol, self).__init__()
        self._gamma_h = 267.5987E6 # radians s^-1 T^-1 (s = seconds, T = Tesla)
        self._unweighted_threshold = 25e6 # s/m^2
        self._columns = {}
        self._preferred_column_order = ('gx', 'gy', 'gz', 'G', 'Delta', 'delta', 'TE', 'T1', 'b', 'q', 'maxG')
        self._virtual_columns = [VirtualColumnB(),
                                 SimpleVirtualColumn('Delta', lambda protocol: get_sequence_timings(protocol)['Delta']),
                                 SimpleVirtualColumn('delta', lambda protocol: get_sequence_timings(protocol)['delta']),
                                 SimpleVirtualColumn('G', lambda protocol: get_sequence_timings(protocol)['G'])]

        if columns:
            self._columns = columns

            for k, v in columns.items():
                s = v.shape
                if len(s) > 2 or (len(s) == 2 and s[1] > 1):
                    raise ValueError("All columns should be of width one.")

                if len(s) ==2 and s[1] == 1:
                    self._columns[k] = np.squeeze(v)

    @property
    def gamma_h(self):
        """Get the used gamma of the ``H`` atom used by this protocol.

        Returns:
            float: The used gamma of the ``H`` atom used by this protocol.
        """
        return self._gamma_h

    def update_column(self, name, data):
        """Updates the given column with new information.

        This actually calls add_column(name, data) in the background. This function is included to allow for clearer
        function calls.

        Args:
            name (str): The name of the column to add
            data (ndarray): The vector to add to this protocol.

        Returns:
            self: for chaining
        """
        return self.add_column(name, data)

    def add_column(self, name, data):
        """Add a column to this protocol. This overrides the column if present.

        Args:
            name (str): The name of the column to add
            data (ndarray): The vector to add to this protocol.

        Returns:
            self: for chaining
        """
        if isinstance(data, six.string_types):
            data = float(data)

        if isinstance(data, numbers.Number) or not data.shape:
            if self._columns:
                data = np.ones((self.length,)) * data
            else:
                data = np.ones((1,)) * data

        s = data.shape
        if self.length and s[0] > self.length:
            self._logger.info("The column '{}' has to many elements ({}), we will only use the first {}.".format(
                name, s[0], self.length))
            self._columns.update({name: data[:self.length]})
        elif self.length and s[0] < self.length:
            raise ValueError("Incorrect column length given for '{}', expected {} and got {}.".format(
                name, self.length, s[0]))
        else:
            if name == 'g' and len(data.shape) > 1 and data.shape[1] == 3:
                self._columns.update({'gx': data[:, 0], 'gy': data[:, 1], 'gz': data[:, 2]})
            else:
                self._columns.update({name: data})
        return self

    def add_column_from_file(self, name, file_name, multiplication_factor=1):
        """Add a column to this protocol, loaded from the given file.

        The given file can either contain a single value (which is broadcasted), or one value per protocol line.

        Args:
            name (str): The name of the column to add
            file_name (str): The file to get the column from.
            multiplication_factor (double): we might need to scale the data by a constant. For example,
                if the data in the file is in ms we might need to scale it to seconds by multiplying with 1e-3
        Returns:
            self: for chaining
        """
        if name == 'g':
            self._columns.update(get_g_columns(file_name))
            for column_name in ('gx', 'gy', 'gz'):
                self._columns[column_name] *= multiplication_factor
            return self
        data = np.genfromtxt(file_name)
        data *= multiplication_factor
        self.add_column(name, data)
        return self

    def remove_column(self, column_name):
        """Completely remove a column from this protocol.

        Args:
            column_name (str): The name of the column to remove
        """
        if column_name == 'g':
            del self._columns['gx']
            del self._columns['gy']
            del self._columns['gz']
        else:
            if column_name in self._columns:
                del self._columns[column_name]

    def remove_rows(self, rows):
        """Remove a list of rows from all the columns.

        Please note that the protocol is 0 indexed.

        Args:
            rows (list of int): List with indices of the rows to remove
        """
        for key, column in self._columns.items():
            self._columns[key] = np.delete(column, rows)

    def get_columns(self, column_names):
        """Get a matrix containing the requested column names in the order given.

        Returns:
            ndarrray: A 2d matrix with the column requested concatenated.
        """
        if not column_names:
            return None
        return np.concatenate([self[i] for i in column_names], axis=1)

    def get_column(self, column_name):
        """Get the column associated by the given column name.

        Args:
            column_name (str): The name of the column we want to return.

        Returns:
            ndarray: The column we would like to return. This is returned as a 2d matrix with shape (n, 1).

        Raises:
            KeyError: If the column could not be found.
        """
        try:
            return self._get_real_column(column_name)
        except KeyError:
            try:
                return self._get_estimated_column(column_name)
            except KeyError:
                raise KeyError('The given column name "{}" could not be found in this protocol.'.format(column_name))

    @property
    def length(self):
        """Get the length of this protocol.

        Returns:
            int: The length of the protocol.
        """
        if self._columns:
            return self._columns[list(self._columns.keys())[0]].shape[0]
        return 0

    @property
    def number_of_columns(self):
        """Get the number of columns in this protocol.

        This only counts the real columns, not the estimated ones.

        Returns:
            int: The number columns in this protocol.
        """
        return len(self._columns)

    @property
    def column_names(self):
        """Get the names of the columns.

        This only lists the real columns, not the estimated ones.

        Returns:
            list of str: The names of the columns.
        """
        return self._column_names_in_preferred_order(self._columns.keys())

    @property
    def estimated_column_names(self):
        """Get the names of the virtual columns.

        This will only return the names of the virtual columns for which no real column exists.
        """
        return self._column_names_in_preferred_order([e.name for e in self._virtual_columns if
                                                      e.name not in self.column_names])

    def get_nmr_shells(self):
        """Get the number of unique shells in this protocol.

        This is measured by counting the number of unique weighted bvals in this protocol.

        Returns:
            int: The number of unique weighted b-values in this protocol

        Raises:
            KeyError: This function may throw a key error if the 'b' column in the protocol could not be loaded.
        """
        return len(self.get_b_values_shells())

    def get_b_values_shells(self):
        """Get the b-values of the unique shells in this protocol.

        Returns:
            :class:`list`: a list with the unique weighted bvals in this protocol.

        Raises:
            KeyError: This function may throw a key error if the 'b' column in the protocol could not be loaded.
        """
        return np.unique(self.get_column('b')[self.get_weighted_indices()]).tolist()

    def count_occurences(self, column, value):
        """Count the occurences of the given value in the given column.

        This can for example be used to count the occurences of a single b-value in the protocol.

        Args:
            column (str): the name of the column
            value (float): the value to count for occurences
        """
        return sum(1 for v in self[column] if v == value)

    def has_column(self, column_name):
        """Check if this protocol has a column with the given name.

        This will also return true if the column can be estimated from the other columns. See is_column_real()
        to get information for columns that are really known.

        Returns:
            boolean: true if there is a column with the given name, false otherwise.
        """
        try:
            return self.get_column(column_name) is not None
        except KeyError:
            return False

    def is_column_real(self, column_name):
        """Check if this protocol has real column information for the column with the given name.

        For example, the other function has_column('G') will normally return true since 'G' can be estimated from 'b'.
        This function will return false if the column needs to be estimated and will return true if real data
        is available for the columnn.

        Returns:
            boolean: true if there is really a column with the given name, false otherwise.
        """
        return column_name in self._columns

    def get_unweighted_indices(self, unweighted_threshold=None):
        """Get the indices to the unweighted volumes.

        If the column 'b' could not be found, assume that all measurements are unweighted.

        Args:
            unweighted_threshold (float): the threshold under which we call it unweighted.

        Returns:
            list of int: A list of indices to the unweighted volumes.
        """
        unweighted_threshold = unweighted_threshold or self._unweighted_threshold

        try:
            b = self.get_column('b')
            g = self.get_column('g')

            g_limit = np.sqrt(g[:, 0]**2 + g[:, 1]**2 + g[:, 2]**2) < 0.99
            b_limit = b[:, 0] < unweighted_threshold

            return np.where(g_limit + b_limit)[0]
        except KeyError:
            return range(self.length)

    def get_weighted_indices(self, unweighted_threshold=None):
        """Get the indices to the weighted volumes.

        Args:
            unweighted_threshold (float): the threshold under which we call it unweighted.

        Returns:
            list of int: A list of indices to the weighted volumes.
        """
        return sorted(set(range(self.get_column('b').shape[0])) -
                      set(self.get_unweighted_indices(unweighted_threshold=unweighted_threshold)))

    def get_indices_bval_in_range(self, start=0, end=1.0e9):
        """Get the indices of the b-values in the range [start, end].

        This can be used to get the indices of gradients whose b-value is in the range suitable for
        a specific analysis.

        Note that we use SI units and you need to specify the values in units of s/m^2 and not in s/mm^2.

        Also note that specifying 0 as start of the range does not automatically mean that the unweighted volumes are
        returned. It can happen that the b-value of the unweighted volumes is higher then 0 even if the the gradient
        ``g`` is ``[0 0 0]``. This function does not make any assumptions about that and just returns indices in the
        given range.

        If you want to include the unweighted volumes, make a call to :meth:`get_unweighted_indices` yourself.

        Args:
            start (float): b-value of the start of the range (inclusive) we want to get the indices of the volumes from.
                Should be positive. We subtract epsilon for float comparison
            end (float): b-value of the end of the range (inclusive) we want to get the indices of the volumes from.
                Should be positive. We add epsilon for float comparison
            epsilon (float): the epsilon we use in the range.

        Returns:
            :class:`list`: a list of indices of all volumes whose b-value is in the given range.
                If you want to include the unweighted volumes, make a call to get_unweighted_indices() yourself.
        """
        b_values = self.get_column('b')
        return np.where((start <= b_values) * (b_values <= end))[0]

    def get_all_columns(self):
        """Get all real (known) columns as a big array.

        Returns:
            ndarray: All the real columns of this protocol.
        """
        return self.get_columns(self.column_names)

    def deepcopy(self):
        """Return a deep copy of this protocol.

        Returns:
            Protocol: A deep copy of this protocol.
        """
        return Protocol(columns=copy.deepcopy(self._columns))

    def append_protocol(self, protocol):
        """Append another protocol to this protocol.

        This will add the columns of the other protocol to the columns of this protocol. This supposes both protocols
        have the same columns.
        """
        if type(protocol) is type(self):
            for key, value in self._columns.items():
                self._columns[key] = np.append(self[key], protocol[key], 0)

    def get_new_protocol_with_indices(self, indices):
        """Create a new protocol object with all the columns but as rows only those of the given indices.

        Args:
            indices: the indices we want to use in the new protocol

        Returns:
            Protocol: a protocol with all the data of the given indices
        """
        return Protocol(columns={k: v[indices] for k, v in self._columns.items()})

    def _get_real_column(self, column_name):
        """Try to use a real column from this protocol.

        Returns:
            ndarray: A real column, that is, a column from which we have real data.

        Raises:
            KeyError: If the column name could not be found we raise a key error.
        """
        if column_name in self._columns:
            return np.reshape(self._columns[column_name], (-1, 1))

        if column_name == 'g':
            return self.get_columns(('gx', 'gy', 'gz'))

        raise KeyError('The given column could not be found.')

    def _get_estimated_column(self, column_name):
        """Try to use an estimated column from this protocol.

        This uses the list of virtual columns to try to estimate the requested column.

        Returns:
            ndarray: An estimated column, that is, a column we estimate from the other columns.

        Raises:
            KeyError: If the column name could not be estimated we raise a key error.
        """
        for virtual_column in self._virtual_columns:
            if virtual_column.name == column_name:
                return np.reshape(virtual_column.get_values(self), (-1, 1))

        raise KeyError('The given column name "{}" could not be found in this protocol.'.format(column_name))

    def _column_names_in_preferred_order(self, column_names):
        """Sort the given column names in the preferred order.

        Column names not in the list of preferred ordering are appended to the end of the list.

        Args:
            column_names (list): the list of column names

        Returns:
            list: the list of column names in the preferred order
        """
        columns_list = [n for n in column_names]
        final_list = []
        for column_name in self._preferred_column_order:
            if column_name in columns_list:
                columns_list.remove(column_name)
                final_list.append(column_name)

        final_list.extend(columns_list)
        return final_list

    def __len__(self):
        return self.length

    def __contains__(self, column):
        return self.has_column(column)

    def __getitem__(self, column):
        return self.get_column(column)

    def __delitem__(self, key):
        return self.remove_column(key)

    def __iter__(self):
        for key in self._columns.keys():
            yield key

    def __setitem__(self, key, value):
        return self.add_column(key, value)

    def __str__(self):
        s = 'Column names: ' + ', '.join(self.column_names) + "\n"
        s += 'Data: ' + "\n"
        s += np.array_str(self.get_all_columns())
        return s


class VirtualColumn(object):

    def __init__(self, name):
        """The interface for generating virtual columns.

        Virtual columns are columns generated on the fly from the other parts of the protocol. They are
        generally only generated if the column it tries to generate is not in the protocol.

        In the Protocol they are used separately from the RealColumns. The VirtualColumns can always be added to
        the Protocol, but are only used when needed. The RealColumns can overrule VirtualColumns by their presence.

        Args:
            name (str): the name of the column this object generates.
        """
        self.name = name

    def get_values(self, parent_protocol):
        """Get the column given the information in the given protocol.

        Args:
            parent_protocol (Protocol): the protocol object to use as a basis for generating the column

        Returns:
            ndarray: the single column as a row vector or 2d matrix of shape nx1
        """


class SimpleVirtualColumn(VirtualColumn):

    def __init__(self, name, generate_function):
        """Create a simple virtual column that uses the given generate function to get the column.

        Args:
            name (str): the name of the column
            generate_function (python function): the function to generate the column
        """
        super(SimpleVirtualColumn, self).__init__(name)
        self._generate_function = generate_function

    def get_values(self, parent_protocol):
        return self._generate_function(parent_protocol)


class VirtualColumnB(VirtualColumn):

    def __init__(self):
        super(VirtualColumnB, self).__init__('b')

    def get_values(self, parent_protocol):
        sequence_timings = get_sequence_timings(parent_protocol)
        return np.reshape(np.array(parent_protocol.gamma_h ** 2 *
                                   sequence_timings['G'] ** 2 *
                                   sequence_timings['delta'] ** 2 *
                                   (sequence_timings['Delta'] - (sequence_timings['delta'] / 3))), (-1, 1))


def get_sequence_timings(protocol):
    """Return G, Delta and delta, estimate them if necessary.

    If Delta and delta are available, they are used instead of estimated Delta and delta.

    Args:
        protocol (Protocol): the protocol for which we want to get the sequence timings.

    Returns:
        dict: the columns G, Delta and delta
    """
    def all_real(columns):
        return all(map(protocol.is_column_real, columns))

    if all_real(['G', 'delta', 'Delta']):
        return {name: protocol[name] for name in ['G', 'delta', 'Delta']}

    if all_real(['b', 'Delta', 'delta']):
        G = np.sqrt(protocol['b'] / (protocol.gamma_h ** 2 * protocol['delta'] ** 2 *
                                     (protocol['Delta'] - (protocol['delta'] / 3.0))))
        G[protocol.get_unweighted_indices()] = 0
        return {'G': G, 'Delta': protocol['Delta'], 'delta': protocol['delta']}

    if all_real(['b', 'Delta', 'G']):
        input_array = np.zeros((protocol.length, 4))
        input_array[:, 0] = -1 / 3.0
        input_array[:, 1] = np.squeeze(protocol['Delta'])
        input_array[:, 2] = 0
        input_array[:, 3] = np.squeeze(-protocol['b'] / (protocol.gamma_h ** 2 * protocol['G'] ** 2))

        b = protocol['b']
        delta = np.zeros((protocol.length, 1))

        for ind in range(protocol.length):
            if b[ind] == 0:
                delta[ind] = 0
            else:
                roots = np.roots(input_array[ind])
                delta[ind] = roots[0]

        return {'G': protocol['G'], 'Delta': protocol['Delta'], 'delta': delta}

    if all_real(['b', 'G', 'delta']):
        Delta = np.nan_to_num(np.array((protocol['b'] - protocol.gamma_h ** 2 * protocol['G'] ** 2
                                        * protocol['delta'] ** 3 / 3.0) /
                                       (protocol.gamma_h ** 2 * protocol['G'] ** 2 * protocol['delta'] ** 2)))
        return {'G': protocol['G'], 'delta': protocol['delta'], 'Delta': Delta}

    if not protocol.is_column_real('b'):
        raise KeyError('Can not estimate the sequence timings, column "b" is not provided.')

    if protocol.has_column('maxG'):
        maxG = protocol['maxG']
    else:
        maxG = np.reshape(np.ones((protocol.length,)) * 0.04, (-1, 1))

    bvals = protocol['b']
    if protocol.get_b_values_shells():
        bmax = max(protocol.get_b_values_shells())
    else:
        bmax = 1

    Deltas = (3 * bmax / (2 * protocol.gamma_h ** 2 * maxG ** 2)) ** (1 / 3.0)
    deltas = Deltas
    G = np.sqrt(bvals / bmax) * maxG

    return {'G': G, 'Delta': Deltas, 'delta': deltas}


def load_bvec_bval(bvec, bval, column_based='auto', bval_scale='auto'):
    """Load an protocol from a bvec and bval file.

    This supposes that the bvec (the vector file) has 3 rows (gx, gy, gz) and is space or tab seperated.
    The bval file (the b values) are one one single line with space or tab separated b values.

    Args:
        bvec (str): The filename of the bvec file
        bval (str): The filename of the bval file
        column_based (boolean): If true, this supposes that the bvec (the vector file) has 3 rows (gx, gy, gz)
            and is space or tab seperated and that the bval file (the b values) are one one single line with space or
            tab separated b values. If false, the vectors and b values are each one a different line.
            If 'auto' it is autodetected, this is the default.
        bval_scale (float): The amount by which we want to scale (multiply) the b-values. The default is auto,
            this checks if the b-val is lower then 1e4 and if so multiplies it by 1e6.
            (sets bval_scale to 1e6 and multiplies), else multiplies by 1.

    Returns:
        Protocol the loaded protocol.
    """
    bvec = get_g_columns(bvec, column_based=column_based)
    bval = np.expand_dims(np.genfromtxt(bval), axis=1)

    if bval_scale == 'auto' and bval[0, 0] < 1e4:
        bval *= 1e6
    else:
        bval *= bval_scale

    columns = {'b': bval}
    columns.update(bvec)

    if bvec['gx'].shape[0] != bval.shape[0]:
        raise ValueError('Columns not of same length.')

    return Protocol(columns=columns)


def get_g_columns(bvec_file, column_based='auto'):
    """Get the columns of a bvec file. Use auto transpose if needed.

    Args:
        bvec_file (str): The filename of the bvec file
        column_based (boolean): If true, this supposes that the bvec (the vector file) has 3 rows (gx, gy, gz)
            and is space or tab seperated
            If false, the vectors are each one a different line.
            If 'auto' it is autodetected, this is the default.

    Returns:
        dict: the loaded bvec matrix separated into 'gx', 'gy' and 'gz'
    """
    bvec = np.genfromtxt(bvec_file)

    if len(bvec.shape) < 2:
        raise ValueError('Bvec file does not have enough dimensions.')

    if column_based == 'auto':
        if bvec.shape[1] > bvec.shape[0]:
            bvec = bvec.transpose()
    elif column_based:
        bvec = bvec.transpose()

    return {'gx': np.reshape(bvec[:, 0], (-1, 1)),
            'gy': np.reshape(bvec[:, 1], (-1, 1)),
            'gz': np.reshape(bvec[:, 2], (-1, 1))}


def write_bvec_bval(protocol, bvec_fname, bval_fname, column_based=True, bval_scale=1):
    """Write the given protocol to bvec and bval files.

    This writes the bvector and bvalues to the given filenames.

    Args:
        protocol (Protocol): The protocol to write to bvec and bval files.
        bvec_fname (string): The bvector filename
        bval_fname (string): The bval filename
        column_based (boolean, optional, default true):
            If true, this supposes that the bvec (the vector file) will have 3 rows (gx, gy, gz)
            and will be space or tab seperated and that the bval file (the b values) are one one single line
            with space or tab separated b values.
        bval_scale (double or str): the amount by which we want to scale (multiply) the b-values.
            The default is auto, this checks if the first b-value is higher than 1e4 and if so multiplies it by
            1e-6 (sets bval_scale to 1e-6 and multiplies), else multiplies by 1.
    """
    b = protocol['b'].copy()
    g = protocol['g'].copy()

    if bval_scale == 'auto':
        if b[0] > 1e4:
            b *= 1e-6
    else:
        b *= bval_scale

    if column_based:
        b = b.transpose()
        g = g.transpose()

    for d in (bvec_fname, bval_fname):
        if not os.path.isdir(os.path.dirname(d)):
            os.makedirs(os.path.dirname(d))

    np.savetxt(bvec_fname, g)
    np.savetxt(bval_fname, b)


def load_protocol(protocol_fname):
    """Load an protocol from the given protocol file, with as column names the given list of names.

    Args:
        protocol_fname (string): The filename of the protocol file to use.
            This should be a comma separated or tab delimited file with equal length columns.

    Returns:
        An protocol with all the columns loaded.
    """
    with open(protocol_fname) as f:
        protocol = f.readlines()

    if protocol[0][0] != '#':
        raise ProtocolIOError('No column names defined in protocol.')

    column_names = [c.strip() for c in protocol[0][1:-1].split(',')]

    data = np.genfromtxt(protocol_fname)
    s = data.shape
    d = {}

    if len(s) == 1:
        d.update({column_names[0]: data})
    else:
        for i in range(s[1]):
            d.update({column_names[i]: data[:, i]})

    return Protocol(columns=d)


def write_protocol(protocol, fname, columns_list=None):
    """Write the given protocol to a file.

    Args:
        protocol (Protocol): The protocol to write to file
        fname (string): The filename to write to
        columns_list (tuple): The tuple with the columns names to write (and in that order).
            If None, all the columns are written to file.

    Returns:
        tuple: the parameters that where written (and in that order)
    """
    if not columns_list:
        columns_list = protocol.column_names

        if 'G' in columns_list and 'Delta' in columns_list and 'delta' in columns_list:
            if 'b' in columns_list:
                columns_list.remove('b')
            if 'maxG' in columns_list:
                columns_list.remove('maxG')

    data = protocol.get_columns(columns_list)

    if not os.path.isdir(os.path.dirname(fname)):
        os.makedirs(os.path.dirname(fname))

    with open(fname, 'w') as f:
        f.write('#')
        f.write(','.join(columns_list))
        f.write("\n")

    with open(fname, 'ab') as f:
        np.savetxt(f, data, delimiter="\t")

    if columns_list:
        return columns_list
    return protocol.column_names


def auto_load_protocol(directory, protocol_columns=None, bvec_fname=None, bval_fname=None, bval_scale='auto'):
    """Load a protocol from the given directory.

    This function will only auto-search files in the top directory and not in the sub-directories.

    This will first try to use the first .prtcl file found. If none present, it will try to find bval and bvec files
    to use and then try to find the protocol options.

    The protocol_options should be a dictionary mapping protocol items to filenames. If given, we only use the items
    in that dictionary. If not given we try to autodetect the protocol option files from the given directory.

    The search order is (continue until matched):

        1) anything ending in .prtcl
        2) a) the given bvec and bval file
           b) anything containing bval or b-val
           c) anything containing bvec or b-vec
                i) This will prefer a bvec file that also has 'fsl' in the name. This to be able to auto use
                    HCP MGH bvec directions.
           d) protocol options
                i) using dict
                ii) matching filenames exactly to the available protocol options.
                    (e.g, finding a file named TE for the TE's)

    The available protocol options are:

        - TE: the TE in seconds, either a file or, one value or one value per bvec
        - TR: the TR in seconds, either a file or, either one value or one value per bvec
        - Delta: the big Delta in seconds, either a file or, either one value or one value per bvec
        - delta: the small delta in seconds, either a file or, either one value or one value per bvec
        - maxG: the maximum gradient amplitude G in T/m. Used in estimating G, Delta and delta if not given.

    Args:
        directory (str): the directory to use the protocol from
        protocol_columns (dict): mapping protocol items to filenames (as a subpath of the given directory)
            or mapping them to values (one value or one value per bvec line)
        bvec_fname (str): if given, the filename of the bvec file (as a subpath of the given directory)
        bval_fname (str): if given, the filename of the bvec file (as a subpath of the given directory)
        bval_scale (double): The scale by which to scale the values in the bval file.
            If we use from bvec and bval we will use this scale. If 'auto' we try to guess the units/scale.

    Returns:
        Protocol: a loaded protocol file.

    Raises:
        ValueError: if not enough information could be found. (No protocol or no bvec/bval combo).
    """
    protocol_files = list(glob.glob(os.path.join(directory, '*.prtcl')))
    if protocol_files:
        return load_protocol(protocol_files[0])

    if not bval_fname:
        bval_files = list(glob.glob(os.path.join(directory, '*bval*')))
        if not bval_files:
            bval_files = glob.glob(os.path.join(directory, '*b-val*'))
            if not bval_files:
                raise ValueError('Could not find a suitable bval file')
        bval_fname = bval_files[0]

    if not bvec_fname:
        bvec_files = list(glob.glob(os.path.join(directory, '*bvec*')))
        if not bvec_files:
            bvec_files = glob.glob(os.path.join(directory, '*b-vec*'))
            if not bvec_files:
                raise ValueError('Could not find a suitable bvec file')

        for bvec_file in bvec_files:
            if 'fsl' in os.path.basename(bvec_file):
                bvec_fname = bvec_files[0]

        if not bvec_fname:
            bvec_fname = bvec_files[0]

    protocol = load_bvec_bval(bvec_fname, bval_fname, bval_scale=bval_scale)

    protocol_extra_cols = ['TE', 'TR', 'Delta', 'delta', 'maxG']

    if protocol_columns:
        for col in protocol_extra_cols:
            if col in protocol_columns:
                if isinstance(protocol_columns[col], six.string_types):
                    protocol.add_column_from_file(col, os.path.join(directory, protocol_columns[col]))
                else:
                    protocol.add_column(col, protocol_columns[col])
    else:
        for col in protocol_extra_cols:
            if os.path.isfile(os.path.join(directory, col)):
                protocol.add_column_from_file(col, os.path.join(directory, col))

    return protocol



from mdt.models.compartments import CompartmentConfig
from mdt.components_loader import bind_function

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CylinderGPD(CompartmentConfig):

    parameter_list = ('g', 'G', 'Delta', 'delta', 'd', 'theta', 'phi', 'R')
    dependency_list = ('MRIConstants',
                       'NeumannCylPerpPGSESum')
    cl_code = '''
        mot_float_type sum = NeumannCylPerpPGSESum(Delta, delta, d, R);

        const mot_float_type4 n = (mot_float_type4)(cos(phi) * sin(theta), sin(phi) * sin(theta), cos(theta), 0.0);
        mot_float_type omega = (G == 0.0) ? M_PI_2 : acos(dot(n, g * G) / (G * length(n)));

        return exp(-2 * GAMMA_H_SQ * pown(G * sin(omega), 2) * sum) *
                exp(-(Delta - (delta/3.0)) * pown(GAMMA_H * delta * G * cos(omega), 2) * d);
    '''

    @bind_function
    def get_extra_results_maps(self, results_dict):
        return self._get_vector_result_maps(results_dict[self.name + '.theta'],
                                            results_dict[self.name + '.phi'])



from mdt.models.compartments import CompartmentConfig
from mdt.components_loader import bind_function
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI_IC(CompartmentConfig):

    parameter_list = ('g', 'b', 'G', 'Delta', 'delta', 'd', 'theta', 'phi', 'kappa', 'R')
    dependency_list = ('CerfErfi',
                       'MRIConstants',
                       'NeumannCylPerpPGSESum')

    @bind_function
    def get_extra_results_maps(self, results_dict):
        maps = self._get_vector_result_maps(results_dict[self.name + '.theta'],
                                            results_dict[self.name + '.phi'])
        maps.update({self.name + '.odi': np.arctan2(1.0, results_dict[self.name + '.kappa'] * 10) * 2 / np.pi})
        return maps



import signal
import sys

from PyQt5.QtCore import pyqtSignal

from mdt.gui.model_fit.design.ui_about_dialog import Ui_AboutDialog
from mdt.gui.model_fit.design.ui_runtime_settings_dialog import Ui_RuntimeSettingsDialog
from mdt.gui.model_fit.tabs.fit_model_tab import FitModelTab
from mdt.gui.model_fit.tabs.generate_brain_mask_tab import GenerateBrainMaskTab
from mdt.gui.model_fit.tabs.generate_roi_mask_tab import GenerateROIMaskTab
from mdt.gui.model_fit.tabs.view_results_tab import ViewResultsTab

import mdt.utils
import mot.configuration
from mdt.configuration import update_gui_config
from mdt.gui.model_fit.tabs.generate_protocol_tab import GenerateProtocolTab
from mot.cl_environments import CLEnvironmentFactory
from mot.load_balance_strategies import EvenDistribution

try:
    #python 2.7
    from Queue import Queue
except ImportError:
    # python 3.4
    from queue import Queue
from PyQt5 import QtGui
from PyQt5.QtCore import QThread, QTimer, pyqtSlot
from PyQt5.QtWidgets import QMainWindow, QDialog, QDialogButtonBox
from mdt.gui.model_fit.design.ui_main_gui import Ui_MainWindow
from mdt.gui.utils import print_welcome_message, ForwardingListener, MessageReceiver, center_window, QtManager
from mdt.gui.model_fit.utils import SharedState
from mdt.log_handlers import LogDispatchHandler

__author__ = 'Robbert Harms'
__date__ = "2016-06-26"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MDTGUISingleModel(QMainWindow, Ui_MainWindow):

    def __init__(self, shared_state, computations_thread):
        super(MDTGUISingleModel, self).__init__()
        self.setupUi(self)
        self._shared_state = shared_state

        self._computations_thread = computations_thread
        self._computations_thread.signal_starting.connect(self.computations_started)
        self._computations_thread.signal_finished.connect(self.computations_finished)

        self._stdout_old = sys.stdout
        self._stderr_old = sys.stderr
        self._logging_update_queue = Queue()
        self._logging_update_thread = QThread()

        self._message_receiver = MessageReceiver(self._logging_update_queue)
        self._message_receiver.text_message_signal.connect(self.update_log)

        self._message_receiver.moveToThread(self._logging_update_thread)
        self._logging_update_thread.started.connect(self._message_receiver.run)
        self._logging_update_thread.start()
        self._connect_output_textbox()

        self.actionExit.setShortcuts(['Ctrl+q', 'Ctrl+w'])

        self.action_RuntimeSettings.triggered.connect(lambda: RuntimeSettingsDialog(self).exec_())
        self.actionAbout.triggered.connect(lambda: AboutDialog(self).exec_())

        self.executionStatusLabel.setText('Idle')
        self.executionStatusIcon.setPixmap(QtGui.QPixmap(":/main_gui/icon_status_red.png"))

        self.fit_model_tab = FitModelTab(shared_state, self._computations_thread)
        self.fit_model_tab.setupUi(self.fitModelTab)

        self.generate_mask_tab = GenerateBrainMaskTab(shared_state, self._computations_thread)
        self.generate_mask_tab.setupUi(self.generateBrainMaskTab)

        self.view_results_tab = ViewResultsTab(shared_state, self._computations_thread)
        self.view_results_tab.setupUi(self.viewResultsTab)

        self.generate_roi_mask_tab = GenerateROIMaskTab(shared_state, self._computations_thread)
        self.generate_roi_mask_tab.setupUi(self.generateROIMaskTab)

        self.generate_protocol_tab = GenerateProtocolTab(shared_state, self._computations_thread)
        self.generate_protocol_tab.setupUi(self.generateProtocolTab)

        self.tabs = [self.fit_model_tab, self.generate_mask_tab, self.generate_roi_mask_tab,
                     self.generate_protocol_tab, self.view_results_tab]

        self.MainTabs.currentChanged.connect(lambda index: self.tabs[index].tab_opened())

    def _connect_output_textbox(self):
        sys.stdout = ForwardingListener(self._logging_update_queue)
        sys.stderr = ForwardingListener(self._logging_update_queue)
        LogDispatchHandler.add_listener(ForwardingListener(self._logging_update_queue))
        print_welcome_message()

    def closeEvent(self, event):
        sys.stdout = self._stdout_old
        sys.stderr = self._stderr_old
        self._message_receiver.is_running = False
        self._logging_update_thread.quit()
        self._logging_update_thread.wait(10)
        super(MDTGUISingleModel, self).closeEvent(event)

    def send_sigint(self, *args):
        self.close()

    @pyqtSlot()
    def computations_started(self):
        self.executionStatusLabel.setText('Computing')
        self.executionStatusIcon.setPixmap(QtGui.QPixmap(":/main_gui/icon_status_green.png"))

    @pyqtSlot()
    def computations_finished(self):
        self.executionStatusLabel.setText('Idle')
        self.executionStatusIcon.setPixmap(QtGui.QPixmap(":/main_gui/icon_status_red.png"))

    @pyqtSlot(str)
    def update_log(self, string):
        sb = self.loggingTextBox.verticalScrollBar()
        scrollbar_position = sb.value()
        autoscroll = scrollbar_position == sb.maximum()
        self.loggingTextBox.moveCursor(QtGui.QTextCursor.End)
        self.loggingTextBox.insertPlainText(string)

        if autoscroll:
            sb.setValue(sb.maximum())
        else:
            sb.setValue(scrollbar_position)


class RuntimeSettingsDialog(Ui_RuntimeSettingsDialog, QDialog):

    def __init__(self, parent):
        super(RuntimeSettingsDialog, self).__init__(parent)
        self.setupUi(self)

        self.all_cl_devices = CLEnvironmentFactory.smart_device_selection()
        self.user_selected_devices = mot.configuration.get_cl_environments()
        self.cldevicesSelection.itemSelectionChanged.connect(self.selection_updated)

        self.cldevicesSelection.insertItems(0, [str(cl_device) for cl_device in self.all_cl_devices])

        load_balancer = mot.configuration.get_load_balancer()
        lb_used_devices = load_balancer.get_used_cl_environments(self.all_cl_devices)

        for ind, device in enumerate(self.all_cl_devices):
            self.cldevicesSelection.item(ind).setSelected(device in self.user_selected_devices
                                                          and device in lb_used_devices)

        self.buttonBox.button(QDialogButtonBox.Ok).clicked.connect(self._update_settings)

    @pyqtSlot()
    def selection_updated(self):
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            any(self.cldevicesSelection.item(ind).isSelected() for ind in range(self.cldevicesSelection.count())))

    def _update_settings(self):
        selection = [ind for ind in range(self.cldevicesSelection.count())
                     if self.cldevicesSelection.item(ind).isSelected()]
        mot.configuration.set_cl_environments([self.all_cl_devices[ind] for ind in selection])
        mot.configuration.set_load_balancer(EvenDistribution())

        update_gui_config({'runtime_settings': {'cl_device_ind': selection}})


class AboutDialog(Ui_AboutDialog, QDialog):

    def __init__(self, parent):
        super(AboutDialog, self).__init__(parent)
        self.setupUi(self)
        self.contentLabel.setText(self.contentLabel.text().replace('{version}', mdt.__version__))


class ComputationsThread(QThread):

    signal_starting = pyqtSignal()
    signal_finished = pyqtSignal()

    def __init__(self, *args, **kwargs):
        """This is the thread handler for the computations.

        When running computations using this thread please connect signals to the starting and finished slot of this
        class. These handlers notify the main window of the computations.
        """
        super(ComputationsThread, self).__init__(*args, **kwargs)

    @pyqtSlot()
    def starting(self):
        self.signal_starting.emit()

    @pyqtSlot()
    def finished(self):
        self.signal_finished.emit()


def start_gui(base_dir=None, app_exec=True):
    """Start the model fitting GUI.

    Args:
        base_dir (str): the starting directory for the file opening actions
        app_exec (boolean): if true we execute the Qt application, set to false to disable.
    """
    try:
        mdt.configuration.load_user_gui()
    except IOError:
        pass

    app = QtManager.get_qt_application_instance()

    state = SharedState()
    state.base_dir = base_dir

    computations_thread = ComputationsThread()
    computations_thread.start()

    # catches the sigint
    timer = QTimer()
    timer.start(500)
    timer.timeout.connect(lambda: None)

    composite_model_gui = MDTGUISingleModel(state, computations_thread)
    signal.signal(signal.SIGINT, composite_model_gui.send_sigint)

    center_window(composite_model_gui)
    composite_model_gui.show()

    QtManager.add_window(composite_model_gui)
    if app_exec:
        QtManager.exec_()


if __name__ == '__main__':
    start_gui()



import glob
import os

from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QFileDialog

from mdt import results_preselection_names
from mdt.nifti import load_nifti
from mdt.visualization.maps.base import DataInfo, MapPlotConfig
from mdt.gui.maps_visualizer.main import start_gui
from mdt.gui.model_fit.design.ui_view_results_tab import Ui_ViewResultsTabContent
from mdt.gui.utils import MainTab
from mdt.utils import split_image_path

__author__ = 'Robbert Harms'
__date__ = "2016-06-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ViewResultsTab(MainTab, Ui_ViewResultsTabContent):

    def __init__(self, shared_state, computations_thread):
        self._shared_state = shared_state
        self._parameter_files = {}
        self._folder = None

    def setupUi(self, ViewResultsTabContent):
        super(ViewResultsTab, self).setupUi(ViewResultsTabContent)

        self.selectFolderButton.clicked.connect(lambda: self._select_folder())

        self.selectedFolderText.textChanged.connect(self.directory_updated)
        self.viewButton.clicked.connect(self.view_maps)
        self.invertSelectionButton.clicked.connect(self.invert_selection)
        self.deselectAllButton.clicked.connect(self.deselect_all)
        self.initialSliceChooser.valueChanged.connect(self._shared_state.set_slice_index)
        self.initialDimensionChooser.valueChanged.connect(self._shared_state.set_dimension_index)
        self.initialSliceChooser.setMaximum(0)

    def open_dir(self, directory):
        self.selectedFolderText.setText(directory)
        self.directory_updated(directory)

    def _select_folder(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedFolderText.text() != '':
            initial_dir = self.selectedFolderText.text()

        folder = QFileDialog().getExistingDirectory(caption='Select directory to view', directory=initial_dir)

        if os.path.isdir(folder):
            self.selectedFolderText.setText(folder)
            self._shared_state.base_dir = folder

    @pyqtSlot(str)
    def directory_updated(self, folder):
        if os.path.isfile(folder):
            folder = os.path.dirname(folder)

        self._folder = folder
        result_files = glob.glob(os.path.join(folder, '*.nii*'))

        def get_name(img_path):
            return split_image_path(os.path.basename(img_path))[1]

        self._parameter_files = {get_name(f): get_name(f) for f in result_files}

        items_list = sorted(self._parameter_files.keys())
        selected_items = results_preselection_names(sorted(self._parameter_files.keys()))

        self.selectMaps.clear()
        self.selectMaps.addItems(items_list)

        for item in [self.selectMaps.item(index) for index in range(self.selectMaps.count())]:
            if item.text() in selected_items:
                item.setSelected(True)

        if items_list:
            shape = load_nifti(result_files[0]).shape
            maximum = shape[self.initialDimensionChooser.value()]
            self.initialSliceChooser.setMaximum(maximum)

            if self.initialSliceChooser.value() == 0 or self.initialSliceChooser.value() >= maximum:
                self.initialSliceChooser.setValue(maximum // 2.0)

            self.maximumIndexLabel.setText(str(maximum))

    def invert_selection(self):
        for item in [self.selectMaps.item(index) for index in range(self.selectMaps.count())]:
            item.setSelected(not item.isSelected())

    def deselect_all(self):
        for item in [self.selectMaps.item(index) for index in range(self.selectMaps.count())]:
            item.setSelected(False)

    def view_maps(self):
        maps_to_show = []
        for item in [self.selectMaps.item(index) for index in range(self.selectMaps.count())]:
            if item.isSelected():
                maps_to_show.append(item.text())

        data = DataInfo.from_dir(self._folder)

        config = MapPlotConfig()
        config.maps_to_show = maps_to_show
        config.dimension = self.initialDimensionChooser.value()
        config.slice_index = self.initialSliceChooser.value()

        start_gui(data=data, config=config, app_exec=False)

    def tab_opened(self):
        if self._shared_state.output_folder != '':
            self.selectedFolderText.setText(self._shared_state.output_folder)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
from mdt.cli_scripts.mdt_gui import GUI

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


"""A shortcut for mdt_gui"""


class GUI_Shortcut(GUI):
    pass


if __name__ == '__main__':
    GUI_Shortcut().start()



import inspect
import os
from copy import deepcopy
import six

from mdt.components_loader import ComponentConfig, ComponentBuilder, ParametersLoader, method_binding_meta, \
    ComponentConfigMeta
from mdt.utils import spherical_to_cartesian
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.cl_functions.parameters import CurrentObservationParam

__author__ = 'Robbert Harms'
__date__ = "2015-12-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRICompartmentModelFunction(ModelFunction):

    def __init__(self, name, cl_function_name, parameter_list, cl_header, cl_code, dependency_list):
        """Create a new dMRI compartment model function.

        Args:
            name (str): the name of this compartment model
            cl_function_name (str): the name of this function in the CL kernel
            parameter_list (list of CLFunctionParameter): the list of the function parameters
            cl_header (str): the code for the CL header
            cl_code (str): the code for the function in CL
            dependency_list (list): the list of functions we depend on inside the kernel
        """
        super(DMRICompartmentModelFunction, self).__init__(name, cl_function_name, parameter_list,
                                                           dependency_list=dependency_list)
        self._cl_header = cl_header
        self._cl_code = cl_code

    def get_cl_header(self):
        inclusion_guard_name = 'DMRICM_' + self.cl_function_name + '_H'
        return '''
            {dependencies}
            #ifndef {inclusion_guard_name}
            #define {inclusion_guard_name}
            {header}
            #endif // {inclusion_guard_name}
        '''.format(dependencies=self._get_cl_dependency_headers(), inclusion_guard_name=inclusion_guard_name,
                   header=self._cl_header)

    def get_cl_code(self):
        inclusion_guard_name = 'DMRICM_' + self.cl_function_name + '_CL'
        return '''
            {dependencies}
            #ifndef {inclusion_guard_name}
            #define {inclusion_guard_name}
            {header}
            #endif // {inclusion_guard_name}
        '''.format(dependencies=self._get_cl_dependency_code(), inclusion_guard_name=inclusion_guard_name,
                   header=self._cl_code)

    def _get_vector_result_maps(self, theta, phi, vector_name='vec0'):
        """Convert spherical coordinates to cartesian vector in 3d

        Args:
            theta (ndarray): the double array with the theta values
            phi (ndarray): the double array with the phi values
            vector_name (str): the name for this vector, the common naming scheme is:
                <model_name>.<vector_name>[_{0,1,2}]

        Returns:
            dict: containing the cartesian vector with the main the fibre direction.
                It returns an element .vec0 and elements vec0_
        """
        cartesian = spherical_to_cartesian(theta, phi)
        extra_dict = {'{}.{}'.format(self.name, vector_name): cartesian}

        for ind in range(3):
            extra_dict.update({'{}.{}_{}'.format(self.name, vector_name, ind): cartesian[:, ind]})

        return extra_dict


def _get_parameters_list(parameter_list):
    """Convert all the parameters in the given parameter list to actual parameter objects.

    Args:
        parameter_list (list): a list containing a mix of either parameter objects or strings. If it is a parameter
            we add a copy of it to the return list. If it is a string we will autoload it.

    Returns:
        list: the list of actual parameter objects
    """
    parameters_loader = ParametersLoader()

    parameters = []
    for item in parameter_list:
        if isinstance(item, six.string_types):
            if item == '_observation':
                parameters.append(CurrentObservationParam())
            else:
                parameters.append(parameters_loader.load(item))
        else:
            parameters.append(deepcopy(item))
    return parameters


def _construct_cl_function_definition(return_type, cl_function_name, parameters):
    """Create the CL function definition for a compartment function.

    This will construct something like (for the Stick model):

    .. code-block:: c

        mot_float_type cmStick(const mot_float_type4 g,
                               const mot_float_type b,
                               const mot_float_type d,
                               const mot_float_type theta,
                               const mot_float_type phi)

    Args:
        return_type (str): the return type
        cl_function_name (str): the name of the function
        parameters (list of CLFunctionParameter): the list of function parameters we use for the arguments

    Returns:
        str: the function definition (only the signature).
    """
    def parameter_str(parameter):
        s = parameter.data_type.cl_type

        if parameter.data_type.pre_data_type_type_qualifiers:
            for qualifier in parameter.data_type.pre_data_type_type_qualifiers:
                s = qualifier + ' ' + s

        if parameter.data_type.address_space_qualifier:
            s = parameter.data_type.address_space_qualifier + ' ' + s

        if parameter.data_type.post_data_type_type_qualifier:
            s += ' ' + parameter.data_type.post_data_type_type_qualifier

        s += ' ' + parameter.name

        return s

    parameters_str = ',\n'.join(parameter_str(parameter) for parameter in parameters)
    return '{return_type} {cl_function_name}({parameters})'.format(return_type=return_type,
                                                                   cl_function_name=cl_function_name,
                                                                   parameters=parameters_str)


class CompartmentConfigMeta(ComponentConfigMeta):

    def __new__(mcs, name, bases, attributes):
        """Extends the default meta class with extra functionality for the compartments.

        This adds the cl_function_name if it is not defined, and creates the correct cl_code and cl_header.
        """
        result = super(CompartmentConfigMeta, mcs).__new__(mcs, name, bases, attributes)

        if 'cl_function_name' not in attributes:
            result.cl_function_name = 'cm{}'.format(name)

        # to prevent the base from loading the initial meta class.
        if any(isinstance(base, CompartmentConfigMeta) for base in bases):
            result.cl_code = mcs._get_cl_code(result, bases, attributes)
            result.cl_header = mcs._get_cl_header(result, bases, attributes)

        return result

    @classmethod
    def _get_cl_code(mcs, result, bases, attributes):
        if 'cl_code' in attributes and attributes['cl_code'] is not None:
            s = _construct_cl_function_definition(
                'mot_float_type', result.cl_function_name, _get_parameters_list(result.parameter_list))
            s += '{\n' + attributes['cl_code'] + '\n}'
            return s

        module_path = os.path.abspath(inspect.getfile(result))
        path = os.path.join(os.path.dirname(module_path), os.path.splitext(os.path.basename(module_path))[0]) + '.cl'
        if os.path.isfile(path):
            with open(path, 'r') as f:
                return f.read()

        for base in bases:
            if hasattr(base, 'cl_code') and base.cl_code is not None:
                return base.cl_code

    @classmethod
    def _get_cl_header(mcs, result, bases, attributes):
        if 'cl_header' in attributes and attributes['cl_header'] is not None:
            return attributes['cl_header']

        module_path = os.path.abspath(inspect.getfile(result))
        path = os.path.join(os.path.dirname(module_path), os.path.splitext(os.path.basename(module_path))[0]) + '.h'
        if os.path.isfile(path):
            with open(path, 'r') as f:
                return f.read()

        for base in bases:
            if hasattr(base, 'cl_header') and base.cl_code is not None:
                return base.cl_header

        return _construct_cl_function_definition('mot_float_type', result.cl_function_name,
                                                 _get_parameters_list(result.parameter_list)) + ';'


class CompartmentConfig(six.with_metaclass(CompartmentConfigMeta, ComponentConfig)):
    """The compartment config to inherit from.

    These configs are loaded on the fly by the CompartmentBuilder.

    All methods you define are automatically bound to the DMRICompartmentModelFunction. Also, to do extra
    initialization you can define a method init. This method is called after object construction to allow
    for additional initialization. Also, this method is not added to the final object.

    Attributes:
        name (str): the name of the model, defaults to the class name
        description (str): model description
        cl_function_name (str): the name of the function in the CL kernel
        parameter_list (list): the list of parameters to use. If a parameter is a string we will
            use it automatically, if not it is supposed to be a CLFunctionParameter
            instance that we append directly.
        cl_header (CLHeaderDefinition): the CL header definition to use. Defaults to CLHeaderFromTemplate.
        cl_code (CLCodeDefinition): the CL code definition to use. Defaults to CLCodeFromAdjacentFile.
        dependency_list (list): the list of functions this function depends on, can contain string which will be
            resolved as library functions.
    """
    name = ''
    description = ''
    cl_function_name = None
    parameter_list = []
    cl_header = None
    cl_code = None
    dependency_list = []


class CompartmentBuildingBase(DMRICompartmentModelFunction):
    """Use this class in super calls if you want to overwrite methods in the inherited compartment configs.

    In python2 super needs a type to be able to do its work. This is the type you can give it to allow
    it to do its work.
    """


class CompartmentBuilder(ComponentBuilder):

    def create_class(self, template):
        """Creates classes with as base class CompartmentBuildingBase

        Args:
            template (CascadeConfig): the compartment config template to use for creating the class with the right init
                settings.
        """
        class AutoCreatedDMRICompartmentModel(method_binding_meta(template, CompartmentBuildingBase)):

            def __init__(self, *args):
                new_args = [template.name,
                            template.cl_function_name,
                            _get_parameters_list(template.parameter_list),
                            template.cl_header,
                            template.cl_code,
                            _resolve_dependencies(template.dependency_list)]

                for ind, already_set_arg in enumerate(args):
                    new_args[ind] = already_set_arg

                super(AutoCreatedDMRICompartmentModel, self).__init__(*new_args)

                if hasattr(template, 'init'):
                    template.init(self)

        return AutoCreatedDMRICompartmentModel


def _resolve_dependencies(dependency_list):
    """Resolve the dependency list such that the result contains all functions.

    Args:
        dependency_list (list): the list of dependencies as given by the user. Elements can either include actual
            instances of :class:`~mot.model_building.cl_functions.base.CLFunction` or strings with the name of the
            component to auto-load.

    Returns:
        list: a new list with the string elements resolved
            as :class:`~mot.model_building.cl_functions.base.LibraryFunction`.
    """
    from mdt.components_loader import LibraryFunctionsLoader

    lib_loader = LibraryFunctionsLoader()
    result = []
    for dependency in dependency_list:
        if isinstance(dependency, six.string_types):
            result.append(lib_loader.load(dependency))
        else:
            result.append(dependency)

    return result



from PyQt5.QtCore import QEvent
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QDoubleSpinBox
from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QPlainTextEdit

from mdt.gui.utils import TimedUpdate


class CollapsablePanel(QFrame):

    def __init__(self, parent=None):
        super(CollapsablePanel, self).__init__(parent)

    def toggle(self):
        content = self.findChild(CollapsablePanelContent)
        if content.isVisible():
            content.hide()
        else:
            content.show()

    def set_collapse(self, collapse):
        content = self.findChild(CollapsablePanelContent)
        if collapse:
            content.hide()
        else:
            content.show()


class CollapsablePanelHeader(QLabel):

    def mousePressEvent(self, QMouseEvent):
        super(CollapsablePanelHeader, self).mousePressEvent(QMouseEvent)
        self.parent().toggle()


class CollapsablePanelContent(QFrame):

    def __init__(self, parent=None):
        super(CollapsablePanelContent, self).__init__(parent)


class TextConfigEditor(QPlainTextEdit):

    new_config = pyqtSignal(str)

    def __init__(self, *args):
        super(TextConfigEditor, self).__init__(*args)
        self._timer = TimedUpdate(self._timer_event)
        self.textChanged.connect(lambda: self._timer.add_delayed_callback(400))

    @pyqtSlot()
    def _timer_event(self):
        self.new_config.emit(self.toPlainText())


class MapsReorderer(QListWidget):

    items_reordered = pyqtSignal()

    def __init__(self, *args):
        super(MapsReorderer, self).__init__(*args)
        self.installEventFilter(self)

    def eventFilter(self, sender, event):
        if event.type() == QEvent.ChildRemoved:
            self.items_reordered.emit()
        return False


class QDoubleSpinBoxDotSeparator(QDoubleSpinBox):

    def __init__(self, *args, **kwargs):
        super(QDoubleSpinBoxDotSeparator, self).__init__(*args, **kwargs)

    def valueFromText(self, text):
        return float(text)

    def textFromValue(self, value):
        return str(value)



import numpy as np
from mdt import create_roi
from mdt.utils import ComplexNoiseStdEstimator
from mdt.exceptions import NoiseStdEstimationNotPossible

__author__ = 'Robbert Harms'
__date__ = "2015-11-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TwoUnweightedVolumes(ComplexNoiseStdEstimator):

    def estimate(self, problem_data, **kwargs):
        """Calculate the standard deviation of the error using the first two unweighted volumes/

        This subtracts the values of the first two unweighted volumes from each other, calculates the std over
        the results and divides that by sqrt(2).

        The method is taken from Camino (http://camino.cs.ucl.ac.uk/index.php?n=Man.Estimatesnr).

        Returns:
            float: single value representing the sigma for the given volume
        """
        unweighted_indices = problem_data.protocol.get_unweighted_indices()
        unweighted_volumes = problem_data.dwi_volume[..., unweighted_indices]

        if len(unweighted_indices) < 2:
            raise NoiseStdEstimationNotPossible('Not enough unweighted volumes for this estimator.')

        diff = unweighted_volumes[..., 0] - unweighted_volumes[..., 1]
        voxel_values = create_roi(diff, problem_data.mask)
        return np.std(voxel_values) / np.sqrt(2)

    def __str__(self):
        return __name__



from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco.Lagos'

"""MPM fitting (Weiskopf, 2016 ESMRMB Workshop)

This fitting is a model published by Helms (2008) and Weiskopf (2011) to determinate biological properties
of the tissue/sample in function *of several images*, which includes T1w, PDw and MTw images. This function is still an
approximation and, if the assumptions of those approximations hold for ex-vivo tissue, then can be used
in this data.
"""


class LinMPM_Fit(CompartmentConfig):

    parameter_list = ('TR', 'flip_angle', 'b1_static', 'T1')
    cl_code = 'return log(flip_angle * b1_static) + log(TR / T1) - log( pown(flip_angle * b1_static, 2) / 2 + ( TR / T1 ) ) ;'



import glob
import os
import mdt
from mdt.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo
from mdt.utils import split_image_path

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'Single directory',
             'description': 'All files in one directory, distinct name per subject.',
             'directory_layout':
'''
For every subject in the directory the base name must be equal to be recognized as
belonging to one subject.

Example directory layout:
    /a.nii.gz
    /a.bval
    /a.bvec

    /b.nii
    /b.prtcl

    /c.nii.gz
    /c.prtcl
    /c_mask.nii.gz

This gives three subjects, 'a', 'b' and 'c' with a Diffusion Weighted Image recognized by the extension .nii(.gz) and
a bvec and bval or protocol file for the protocol information. Subject 'c' also has a brain mask associated.

If there is a file mask.nii(.gz) present in the directory it is used as the default mask for all the subjects
in the single directory that do not have their own mask.

You can provide the noise standard deviation to use using a noise_std file (per subject). This should either
be a .txt file (global noise) or a nifti file (voxel wise noise).

Optional items (all case sensitive):
    /<name>.TE
    /<name>.Delta
    /<name>.delta
    /<name>.noise_std.{txt,nii,nii.gz}
'''}


class SingleDir(SimpleBatchProfile):

    def _get_subjects(self):
        pjoin = mdt.make_path_joiner(self._root_dir)

        files = [os.path.basename(f) for f in glob.glob(pjoin('*'))]
        basenames = sorted(list({split_image_path(f)[1] for f in files}))
        subjects = []

        protocol_options = ['TE', 'TR', 'Delta', 'delta', 'maxG']

        default_mask = None
        if list(glob.glob(pjoin('mask.nii*'))):
            default_mask = list(glob.glob(pjoin('mask.nii*')))[0]

        for basename in basenames:
            dwi_fname = None
            if basename + '.nii' in files:
                dwi_fname = pjoin(basename + '.nii')
            elif basename + '.nii.gz' in files:
                dwi_fname = pjoin(basename + '.nii.gz')
            elif basename + '.hdr' in files and basename + '.img' in files:
                dwi_fname = pjoin(basename + '.hdr')

            noise_std = self._autoload_noise_std(basename, file_path=pjoin(basename + '.noise_std'))

            prtcl_fname = None
            if basename + '.prtcl' in files:
                prtcl_fname = pjoin(basename + '.prtcl')

            bval_fname = None
            if basename + '.bval' in files:
                bval_fname = pjoin(basename + '.bval')

            bvec_fname = None
            if basename + '.bvec' in files:
                bvec_fname = pjoin(basename + '.bvec')

            mask_fname = default_mask
            if basename + '_mask.nii' in files:
                mask_fname = pjoin(basename + '_mask.nii')
            elif basename + '_mask.nii.gz' in files:
                mask_fname = pjoin(basename + '_mask.nii.gz')

            extra_cols_from_file = {}
            for option in protocol_options:
                if basename + '.' + option in files:
                    extra_cols_from_file.update({option: pjoin(basename + '.' + option)})

            if dwi_fname and (prtcl_fname or (bval_fname and bvec_fname)):
                protocol_loader = BatchFitProtocolLoader(
                    pjoin(),
                    protocol_fname=prtcl_fname, bvec_fname=bvec_fname, bval_fname=bval_fname,
                    protocol_columns=extra_cols_from_file)

                output_dir = self._get_subject_output_dir(basename, mask_fname, pjoin(self.output_base_dir, basename))

                subjects.append(SimpleSubjectInfo(basename, dwi_fname, protocol_loader, mask_fname, output_dir,
                                                  noise_std=noise_std))
        return subjects

    def __str__(self):
        return meta_info['title']
